"""Validation functions."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import pathlib
from importlib.machinery import SourceFileLoader

import numpy

from bluepyemodel.evaluation.evaluation import compute_responses
from bluepyemodel.evaluation.evaluation import get_evaluator_from_access_point
from bluepyemodel.tools.utils import are_same_protocol
from bluepyemodel.validation import validation_functions

logger = logging.getLogger(__name__)


def define_validation_function(access_point):
    """Define the validation function based on the validation_function setting. If the settings
    was not specified, validate_max_score is used"""

    validation_function = access_point.pipeline_settings.validation_function

    if validation_function is None or not validation_function:
        logger.warning("Validation function not specified, will use validate_max_score.")
        return validation_functions.validate_max_score

    if isinstance(validation_function, str):
        if validation_function == "max_score":
            validation_function = validation_functions.validate_max_score
        elif validation_function == "mean_score":
            validation_function = validation_functions.validate_mean_score
        else:
            raise ValueError("validation_function must be 'max_score' or 'mean_score'.")

    elif isinstance(validation_function, list) and len(validation_function) == 2:
        # pylint: disable=deprecated-method,no-value-for-parameter
        function_module = SourceFileLoader(
            pathlib.Path(validation_function[0]).stem, validation_function[0]
        ).load_module()
        validation_function = getattr(function_module, validation_function[1])

    if not callable(validation_function):
        raise TypeError("validation_function is not callable nor a list of two strings")

    return validation_function


def compute_scores(model, validation_protocols):
    """Compute the scores of an emodel.

    Args:
        model (EModel): emodel
        validation_protocols (list): list of validation protocols
    """
    model.features = model.evaluator.fitness_calculator.calculate_values(model.responses)
    for key, value in model.features.items():
        if value is not None:
            # turn features from arrays to float to be json serializable
            model.features[key] = float(numpy.nanmean([v for v in value if v is not None]))

    scores = model.evaluator.fitness_calculator.calculate_scores(model.responses)
    for feature_name in scores:
        protocol_name = feature_name.split(".")[0]
        if any(are_same_protocol(p, protocol_name) for p in validation_protocols):
            model.scores_validation[feature_name] = scores[feature_name]
        else:
            model.scores[feature_name] = scores[feature_name]


def validate(access_point, mapper, preselect_for_validation=False):
    """Compute the scores and traces for the optimisation and validation
    protocols and perform validation.

    Args:
        access_point (DataAccessPoint): data access point.
        mapper (map): used to parallelize the evaluation of the
            individual in the population.
        preselect_for_validation (bool): True to not re-run already validated models

    Returns:
        emodels (list): list of emodels.
    """

    cell_evaluator = get_evaluator_from_access_point(
        access_point,
        include_validation_protocols=True,
    )

    emodels = compute_responses(
        access_point,
        cell_evaluator=cell_evaluator,
        map_function=mapper,
        preselect_for_validation=preselect_for_validation,
    )

    if not emodels:
        logger.warning("In validate, no emodels for %s", access_point.emodel_metadata.emodel)
        return []

    validation_function = define_validation_function(access_point)

    logger.info("In validate, %s emodels found to validate.", len(emodels))

    for model in emodels:
        compute_scores(model, access_point.pipeline_settings.validation_protocols)

        # turn bool_ into bool to be json serializable
        model.passed_validation = bool(
            validation_function(
                model,
                access_point.pipeline_settings.validation_threshold,
                False,
            )
        )

        access_point.store_emodel(model)

    return emodels



"""EModelWorkflow class"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

logger = logging.getLogger(__name__)


class EModelWorkflow:
    """Contains the state of the workflow and the configurations needed for the workflow"""

    def __init__(
        self,
        targets_configuration_id,
        pipeline_settings_id,
        emodel_configuration_id,
        fitness_configuration_id=None,
        emodels=None,
        emodel_scripts_id=None,
        state="not launched",
    ):
        """Init

        Args:
            targets_configuration (str): TargetsConfiguration nexus id
            pipeline_settings (str): EModelPipelineSettings nexus id
            emodel_configuration (str): NeuronModelConfiguration id
            fitness_configuration_id (str): FitnessCalculatorConfiguration id
            emodels (list): list of EModel ids
            state (str): can be "not launched", "running" or "done"
        """
        self.targets_configuration_id = targets_configuration_id
        self.pipeline_settings_id = pipeline_settings_id
        self.emodel_configuration_id = emodel_configuration_id
        self.fitness_configuration_id = fitness_configuration_id
        self.emodels = emodels if emodels else []
        self.emodel_scripts_id = emodel_scripts_id if emodel_scripts_id else []
        self.state = state

    def add_emodel_id(self, emodel_id):
        """Add an emodel id to the list of emodels"""
        self.emodels.append(emodel_id)

    def add_emodel_script_id(self, emodel_script_id):
        """Add an emodel id to the list of emodels"""
        self.emodel_scripts_id.append(emodel_script_id)

    def get_configuration_ids(self):
        """Return all configuration id parameters"""
        ids = (
            self.targets_configuration_id,
            self.pipeline_settings_id,
            self.emodel_configuration_id,
        )
        if self.fitness_configuration_id:
            ids += tuple([self.fitness_configuration_id])

        return ids

    def get_related_nexus_ids(self):
        emodels_ids = [{"id": id_, "type": "EModel"} for id_ in self.emodels]
        emodel_scripts_ids = [{"id": id_, "type": "EModelScript"} for id_ in self.emodel_scripts_id]
        generates = emodels_ids + emodel_scripts_ids
        if self.fitness_configuration_id:
            generates.append(
                {
                    "id": self.fitness_configuration_id,
                    "type": "FitnessCalculatorConfiguration",
                }
            )

        has_part = []
        if self.targets_configuration_id:
            has_part.append(
                {
                    "id": self.targets_configuration_id,
                    "type": "ExtractionTargetsConfiguration",
                }
            )
        if self.pipeline_settings_id:
            has_part.append({"id": self.pipeline_settings_id, "type": "EModelPipelineSettings"})
        if self.emodel_configuration_id:
            has_part.append({"id": self.emodel_configuration_id, "type": "EModelConfiguration"})

        ids = {}
        if generates:
            ids["generates"] = generates
        if has_part:
            ids["hasPart"] = has_part

        return ids

    def as_dict(self):
        """Used for the storage of the object"""
        return vars(self)



"""Mechanisms related functions"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import os
import shutil
from pathlib import Path

logger = logging.getLogger(__name__)


NEURON_BUILTIN_MECHANISMS = ["hh", "pas", "fastpas", "extracellular", "capacitance"]


def copy_mechs(mechanism_paths, out_dir):
    """Copy mod files in the designated directory.

    Args:
        mechanism_paths (list): list of the paths to the mod files that
            have to be copied.
        out_dir (str): path to directory to which the mod files should
            be copied.
    """

    if mechanism_paths:
        out_dir = Path(out_dir)
        out_dir.mkdir(parents=True, exist_ok=True)

        for m in mechanism_paths:
            p = Path(m["path"])

            if p.is_file():
                new_p = out_dir / p.name
                shutil.copy(str(p), str(new_p))

            else:
                raise FileNotFoundError(
                    "Cannot copy the .mod files locally because the "
                    f"'mechanism_paths' {p} does not exist."
                )


def delete_compiled_mechanisms():
    """Delete compiled mechanisms."""
    if Path("x86_64").is_dir():
        os.popen("rm -rf x86_64").read()


def compile_mechs(mechanisms_dir):
    """Compile the mechanisms.

    Args:
        mechanisms_dir (str): path to the directory containing the
            mod files to compile.
    """

    path_mechanisms_dir = Path(mechanisms_dir)

    if path_mechanisms_dir.is_dir():
        delete_compiled_mechanisms()
        os.popen(f"nrnivmodl {path_mechanisms_dir}").read()
    else:
        raise FileNotFoundError(
            "Cannot compile the mechanisms because 'mechanisms_dir':"
            f" {path_mechanisms_dir} does not exist."
        )


def compile_mechs_in_emodel_dir(mechanisms_directory):
    """Compile mechanisms in emodel directory.

    Args:
        mechanisms_dir (Path): path to the directory containing the
            mod files to compile.
    """
    # pylint: disable=broad-except
    cwd = os.getcwd()

    try:
        os.chdir(str(mechanisms_directory.parents[0]))
        compile_mechs("./mechanisms")
    except Exception as e:
        logger.exception(e)
    finally:
        os.chdir(cwd)


def copy_and_compile_mechanisms(access_point):
    """Copy mechs if asked, and compile them."""

    if access_point.__class__.__name__ == "NexusAccessPoint":
        # Mechanisms are automatically download by the Nexus API
        # when calling this function
        _ = access_point.get_model_configuration()
        compile_mechs("./mechanisms")


def get_mechanism_currents(mech_file):
    """Parse the mech mod file to get the mechanism ion and non-specific currents if any."""
    ion_currs = []
    nonspecific_currents = []
    ionic_concentrations = []
    with open(mech_file, "r") as f:
        mod_lines = f.readlines()
    for line in mod_lines:
        if "WRITE " in line:
            ion_var_name = line.split("WRITE ")[1].rstrip("\n").split(" ")[0]
            # ion current case
            if ion_var_name[0] == "i":
                ion_currs.append(ion_var_name)
                ionic_concentrations.append(f"{ion_var_name[1:]}i")
            # internal ionic concentration case
            elif ion_var_name[-1] == "i":
                ionic_concentrations.append(ion_var_name)
        elif "NONSPECIFIC_CURRENT" in line:
            var_name = line.split("NONSPECIFIC_CURRENT ")[1].rstrip("\n").split(" ")[0]
            if var_name[0] == "i":
                nonspecific_currents.append(var_name)

    return ion_currs, nonspecific_currents, ionic_concentrations


def get_mechanism_name(mech_file):
    """Parse the mech mod file to get the mechanism suffix or point process to use as name."""
    with open(mech_file, "r") as f:
        mod_lines = f.readlines()
    for line in mod_lines:
        if "SUFFIX " in line:
            suffix = line.split("SUFFIX ")[1].rstrip("\n").split(" ")[0]
            return suffix
        if "POINT_PROCESS" in line:
            point_process = line.split("POINT_PROCESS ")[1].rstrip("\n").split(" ")[0]
            return point_process
    raise RuntimeError(f"Could not find SUFFIX nor POINT_PROCESS in {mech_file}")


def discriminate_by_temp(resources, temperatures):
    """Select sublist of resources with given temperature."""
    if not temperatures:
        return resources
    new_temperatures = temperatures.copy()
    temp = new_temperatures.pop(0)
    tmp_resources = [
        r
        for r in resources
        if hasattr(r, "temperature")
        and hasattr(r.temperature, "value")
        and r.temperature.value == temp
    ]
    if len(tmp_resources) > 0 and len(tmp_resources) < len(resources):
        logger.warning(
            "Discriminating resources based on temperature. "
            "Keeping only resource with temperature == %s C.",
            temp,
        )
        return tmp_resources
    if len(new_temperatures) > 0:
        return discriminate_by_temp(resources, new_temperatures)
    return resources



"""Class eFELFeatureBPEM"""

"""
Copyright 2023-2024, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import math

import numpy
from bluepyopt.ephys.efeatures import eFELFeature
from scipy import optimize as opt

from bluepyemodel.tools.multiprotocols_efeatures_utils import get_distances_from_recording_name
from bluepyemodel.tools.multiprotocols_efeatures_utils import get_locations_from_recording_name
from bluepyemodel.tools.multiprotocols_efeatures_utils import get_protocol_list_from_recording_name

logger = logging.getLogger(__name__)


class eFELFeatureBPEM(eFELFeature):
    """eFEL feature extra"""

    SERIALIZED_FIELDS = (
        "name",
        "efel_feature_name",
        "recording_names",
        "stim_start",
        "stim_end",
        "exp_mean",
        "exp_std",
        "threshold",
        "comment",
    )
    # pylint: disable=too-many-arguments

    def __init__(
        self,
        name,
        efel_feature_name=None,
        recording_names=None,
        stim_start=None,
        stim_end=None,
        exp_mean=None,
        exp_std=None,
        threshold=None,
        stimulus_current=None,
        comment="",
        interp_step=None,
        double_settings=None,
        int_settings=None,
        string_settings=None,
        weight=1.0,
    ):
        """Constructor

        Args:
            name (str): name of the eFELFeature object
            efel_feature_name (str): name of the eFeature in the eFEL library
                (ex: 'AP1_peak')
            recording_names (dict): eFEL features can accept several recordings
                as input
            stim_start (float): stimulation start time (ms)
            stim_end (float): stimulation end time (ms)
            exp_mean (float): experimental mean of this eFeature
            exp_std(float): experimental standard deviation of this eFeature
            threshold(float): spike detection threshold (mV)
            comment (str): comment
            weight (float): weight of the efeature.
                Basically multiplies the score of the efeature by this value.
        """

        super().__init__(
            name,
            efel_feature_name,
            recording_names,
            stim_start,
            stim_end,
            exp_mean,
            exp_std,
            threshold,
            stimulus_current,
            comment,
            interp_step,
            double_settings,
            int_settings,
            string_settings,
            max_score=250.0,
        )
        self.weight = weight  # used in objective

    def calculate_bpo_feature(self, responses):
        """Return internal feature which is directly passed as a response"""

        if self.efel_feature_name not in responses:
            return None

        return responses[self.efel_feature_name]

    def calculate_bpo_score(self, responses):
        """Return score for bpo feature"""

        feature_value = self.calculate_bpo_feature(responses)

        if feature_value is None:
            return self.max_score

        return abs(feature_value - self.exp_mean) / self.exp_std

    def _construct_efel_trace(self, responses):
        """Construct trace that can be passed to eFEL"""

        trace = {}
        if "" not in self.recording_names:
            raise ValueError("eFELFeature: '' needs to be in recording_names")
        for location_name, recording_name in self.recording_names.items():
            if location_name == "":
                postfix = ""
            else:
                postfix = f";{location_name}"

            if recording_name not in responses:
                logger.debug(
                    "Recording named %s not found in responses %s",
                    recording_name,
                    str(responses),
                )
                return None

            if responses[self.recording_names[""]] is None or responses[recording_name] is None:
                return None
            trace[f"T{postfix}"] = responses[self.recording_names[""]]["time"]
            trace[f"V{postfix}"] = responses[recording_name]["voltage"]

            if callable(self.stim_start):
                trace[f"stim_start{postfix}"] = [self.stim_start()]
            else:
                trace[f"stim_start{postfix}"] = [self.stim_start]

            if callable(self.stim_end):
                trace[f"stim_end{postfix}"] = [self.stim_end()]
            else:
                trace[f"stim_end{postfix}"] = [self.stim_end]

        return trace

    def _setup_efel(self):
        """Set up efel before extracting the feature"""

        import efel

        efel.reset()

        if self.threshold is not None:
            efel.set_threshold(self.threshold)

        if self.stimulus_current is not None:
            if callable(self.stimulus_current):
                efel.set_double_setting("stimulus_current", self.stimulus_current())
            else:
                efel.set_double_setting("stimulus_current", self.stimulus_current)

        if self.interp_step is not None:
            efel.set_double_setting("interp_step", self.interp_step)

        if self.double_settings is not None:
            for setting_name, setting_value in self.double_settings.items():
                efel.set_double_setting(setting_name, setting_value)

        if self.int_settings is not None:
            for setting_name, setting_value in self.int_settings.items():
                efel.set_int_setting(setting_name, setting_value)

        if self.string_settings is not None:
            for setting_name, setting_value in self.string_settings.items():
                efel.set_str_setting(setting_name, setting_value)

    def calculate_feature(self, responses, raise_warnings=False):
        """Calculate feature value"""
        if self.efel_feature_name.startswith("bpo_"):
            feature_values = numpy.array([self.calculate_bpo_feature(responses)])

        else:
            efel_trace = self._construct_efel_trace(responses)

            if efel_trace is None:
                feature_values = None
            else:
                self._setup_efel()
                logger.debug("Amplitude for %s: %s", self.name, self.stimulus_current)
                import efel

                values = efel.get_feature_values(
                    [efel_trace],
                    [self.efel_feature_name],
                    raise_warnings=raise_warnings,
                )

                feature_values = values[0][self.efel_feature_name]

                efel.reset()

        logger.debug("Calculated values for %s: %s", self.name, str(feature_values))
        return feature_values

    def calulate_score_(self, responses):
        """Calculate the score for non-bpo feature"""
        if self.exp_mean is None:
            score = 0

        else:
            feature_values = self.calculate_feature(responses)
            if (feature_values is None) or (len(feature_values) == 0):
                score = self.max_score
            else:
                score = (
                    numpy.sum(numpy.fabs(feature_values - self.exp_mean))
                    / self.exp_std
                    / len(feature_values)
                )
                logger.debug("Calculated score for %s: %f", self.name, score)

            score = numpy.min([score, self.max_score])
        return score

    def calculate_score(self, responses, trace_check=False):
        """Calculate the score"""

        if self.efel_feature_name.startswith("bpo_"):
            score = self.calculate_bpo_score(responses)

        else:
            score = self.calulate_score_(responses)

        if score is None or math.isnan(score):
            return self.max_score

        return score


class DendFitFeature(eFELFeatureBPEM):
    """Fit to back propagation feature

    To use this class:
        - have "dendrite_backpropagation_fit" as the efeature name
        - have "maximum_voltage_from_voltagebase" as the efel_feature_name
        - have keys in recording names matching the distance from soma, and "" for soma, e.g.
            {"": "soma.v", "50": "dend50.v", "100": "dend100.v", "150": "dend150.v"}
        - have appropriate recordings in protocols
    """

    def __init__(
        self,
        name,
        efel_feature_name=None,
        recording_names=None,
        stim_start=None,
        stim_end=None,
        exp_mean=None,
        exp_std=None,
        threshold=None,
        stimulus_current=None,
        comment="",
        interp_step=None,
        double_settings=None,
        int_settings=None,
        string_settings=None,
        decay=None,
        linear=None,
        weight=1.0,
    ):
        """Constructor"""
        # pylint: disable=too-many-arguments
        super().__init__(
            name,
            efel_feature_name,
            recording_names,
            stim_start,
            stim_end,
            exp_mean,
            exp_std,
            threshold,
            stimulus_current,
            comment,
            interp_step,
            double_settings,
            int_settings,
            string_settings,
            weight,
        )
        self.decay = decay
        self.linear = linear
        self.ymult = None  # set on the fly before the fitting

    @property
    def recording_names_list(self):
        return self.recording_names.values()

    @property
    def distances(self):
        # expects keys in recordings names to be distances from soma (e.g. "50") or "" if at soma
        return [int(rec_name) if rec_name != "" else 0 for rec_name in self.recording_names.keys()]

    @property
    def locations(self):
        # locations used in bpo features (holding current, threshold current)
        # as such, they are implemented for MultiProtocols, but not for simple protocol
        raise NotImplementedError

    def _construct_efel_trace(self, responses):
        """Construct trace that can be passed to eFEL"""

        traces = []
        if "" not in self.recording_names:
            raise ValueError("eFELFeature: '' needs to be in recording_names")

        for recording_name in self.recording_names_list:
            if recording_name not in responses:
                logger.debug(
                    "Recording named %s not found in responses %s",
                    recording_name,
                    str(responses),
                )
                return None

            if responses[recording_name] is None:
                logger.debug("resp of %s is None", responses[recording_name])
                return None

            trace = {}
            trace["T"] = responses[recording_name]["time"]
            trace["V"] = responses[recording_name]["voltage"]
            if callable(self.stim_start):
                trace["stim_start"] = [self.stim_start()]
            else:
                trace["stim_start"] = [self.stim_start]

            if callable(self.stim_end):
                trace["stim_end"] = [self.stim_end()]
            else:
                trace["stim_end"] = [self.stim_end]
            traces.append(trace)

        return traces

    def exp_decay(self, x, p):
        return numpy.exp(-x / p) * self.ymult

    def exp(self, x, p):
        return numpy.exp(x / p) * self.ymult

    def linear_fit(self, x, p):
        return self.ymult + p * x

    def fit(self, distances, values):
        """Fit back propagation"""
        guess = [50]
        self.ymult = values[distances.index(0)]
        if self.linear:
            params, _ = opt.curve_fit(self.linear_fit, distances, values, p0=guess)
        elif self.decay:
            params, _ = opt.curve_fit(self.exp_decay, distances, values, p0=guess)
        else:
            params, _ = opt.curve_fit(self.exp, distances, values, p0=guess)

        return params[0]

    def get_distances_feature_values(self, responses, raise_warnings=False):
        """Compute feature at each distance, and return distances and feature values."""
        distances = []
        feature_values_ = []
        if self.efel_feature_name.startswith("bpo_"):
            feature_names = [
                (f"{self.efel_feature_name}_{loc}" if loc != "soma" else self.efel_feature_name)
                for loc in self.locations
            ]
            feature_values_ = [
                responses[fname]
                for fname in feature_names
                if fname in responses and responses[fname] is not None
            ]
            distances = [
                d
                for d, fname in zip(self.distances, feature_names)
                if fname in responses and responses[fname] is not None
            ]

            # adjust soma value with holding current
            if (
                self.efel_feature_name == "bpo_threshold_current"
                and distances[0] == 0
                and "bpo_holding_current" in responses
                and responses["bpo_holding_current"] is not None
            ):
                feature_values_[0] += responses["bpo_holding_current"]

        else:
            efel_traces = self._construct_efel_trace(responses)

            if efel_traces is None:
                feature_values_ = None
            else:
                self._setup_efel()
                import efel

                values = efel.get_feature_values(
                    efel_traces, [self.efel_feature_name], raise_warnings=raise_warnings
                )
                feature_values_ = [
                    val[self.efel_feature_name][0]
                    for val in values
                    if val[self.efel_feature_name] is not None
                ]
                distances = [
                    d
                    for d, v in zip(self.distances, values)
                    if v[self.efel_feature_name] is not None
                ]

                efel.reset()

        return distances, feature_values_

    def calculate_feature(self, responses, raise_warnings=False):
        """Calculate feature value"""
        distances, feature_values_ = self.get_distances_feature_values(responses, raise_warnings)

        if distances and feature_values_:
            if 0 in distances:
                feature_values = numpy.array([self.fit(distances, feature_values_)])
            else:
                feature_values = None
        else:
            feature_values = None

        logger.debug("Calculated values for %s: %s", self.name, str(feature_values))
        return feature_values

    def calculate_score(self, responses, trace_check=False):
        """Calculate score. bpo and non-bpo feature should use calulate_feature"""
        score = self.calulate_score_(responses)

        if score is None or math.isnan(score):
            return self.max_score

        return score


class DendFitMultiProtocolsFeature(DendFitFeature):
    """Fit across apical dendrite using multiple protocols.

    Attention! Since this feature depends on multiple protocols,
    the stimulus_current passed can be wrong for some of them, and in such a case,
    efel features depending on stimulus_current should not be used.

    To use this class:
        - have a protocol_name with distances in brackets in it
            e.g. LocalInjectionIDrestapic[050,100,150,200]_100
        - same with recording_name
            e.g. apical[055,080,110,200,340].v
        - have corresponding protocols with normal names under "protocols" in config
            e.g. LocalInjectionIDrestapic050_100, LocalInjectionIDrestapic100_100, etc.
        - have a soma protocol under "protocols" with same ecode in config
            e.g. IDrest_100
    """

    @property
    def recording_names_list(self):
        return get_protocol_list_from_recording_name(self.recording_names[""])

    @property
    def distances(self):
        return get_distances_from_recording_name(self.recording_names[""])

    @property
    def locations(self):
        return get_locations_from_recording_name(self.recording_names[""])



"""Distribution Configuration"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class DistributionConfiguration:
    """Contains all the information related to the definition and configuration of a parameter
    distribution"""

    def __init__(
        self,
        name,
        function=None,
        parameters=None,
        morphology_dependent_parameters=None,
        soma_ref_location=0.5,
        comment=None,
    ):
        """Init

        Args:
            name (str): name of the distribution.
            function (str): python function of the distribution as a string. Will be executed
                using the python "eval" method. The string needs to include "value" which will be
                replaced by the conductance of the parameter using the present distribution.
                It can also include "distance" if the distribution is parametrized by the distance
                to the soma. Example: "(-0.8696 + 2.087*math.exp(({distance})*0.0031))*{value}".
            parameters (list of str): names of the parameters that parametrize the above function
                if any. Note that "value" and "distance" do not need to be specified here.
            morphology_dependent_parameters (list of str): unused. To be deprecated.
            soma_ref_location (float): location along the soma used as origin from which to
                compute the distances. Expressed as a fraction (between 0.0 and 1.0).
            comment (str): additional comment or note.
        """

        self.name = name
        self.function = function

        if parameters is None:
            self.parameters = []
        elif isinstance(parameters, str):
            self.parameters = [parameters]
        else:
            self.parameters = parameters

        if morphology_dependent_parameters is None:
            self.morphology_dependent_parameters = []
        elif isinstance(morphology_dependent_parameters, str):
            self.morphology_dependent_parameters = [morphology_dependent_parameters]
        else:
            self.morphology_dependent_parameters = morphology_dependent_parameters

        if soma_ref_location is None:
            soma_ref_location = 0.5
        self.soma_ref_location = soma_ref_location

        self.comment = comment

    def as_dict(self):
        distr_dict = {
            "name": self.name,
            "function": self.function,
            "soma_ref_location": self.soma_ref_location,
        }

        for attr in ["parameters", "morphology_dependent_parameters", "comment"]:
            if getattr(self, attr):
                distr_dict[attr] = getattr(self, attr)

        return distr_dict

    def as_legacy_dict(self):
        distr_dict = {"fun": self.function}

        if self.parameters:
            distr_dict["parameters"] = self.parameters

        return distr_dict



"""Model configuration related functions"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from bluepyemodel.model.model_configurator import ModelConfigurator

logger = logging.getLogger(__name__)


def configure_model(
    access_point,
    morphology_name,
    morphology_path=None,
    morphology_format=None,
    use_gene_data=True,
):
    """Task of creating and saving a neuron model configuration.

    Args:
        access_point (DataAccessPoint): access point to the emodel data
        morphology_name (str): name of the morphology on which to build the neuron model. This
            morphology has to be available in the directory "./morphology" if using the local
            access point or on Nexus if using the Nexus access point.
        morphology_path (str): path to the morphology file
        morphology_format (str): format of the morphology, can be 'asc' or 'swc'. Optional if
            morphology_path was provided.
        use_gene_data (bool): should the configuration be initialized using gene data. If False,
            the configuration will be empty.
    """

    configurator = ModelConfigurator(access_point=access_point)
    configurator.new_configuration(use_gene_data=use_gene_data)

    configurator.configuration.select_morphology(
        morphology_name,
        morphology_path=morphology_path,
        morphology_format=morphology_format,
    )

    configurator.save_configuration()

    return configurator.configuration



"""Morphology related functions"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import pathlib

logger = logging.getLogger(__name__)


def name_morphology(radius, type_="cylindrical"):
    return f"{type_}_morphology_{radius:.4f}"


def cylindrical_morphology_generator(
    radius=5.0, radius_axon=0.0, length_axon=0.0, output_dir="morphologies"
):
    """Creates a cylindrical morphology and save it as a .swc"""

    name = name_morphology(radius, type_="cylindrical")
    morphology_path = pathlib.Path(output_dir) / f"{name}.swc"
    morphology_path.parent.mkdir(parents=True, exist_ok=True)

    content = f"# Cylindrical morphology of radius {radius}\n"
    content += f"1 1 -{radius} 0.0 0.0 {radius} -1\n"
    content += f"2 1 0.0 0.0 0.0 {radius} 1\n"
    content += f"3 1 {radius} 0.0 0.0 {radius} 2\n"

    if radius_axon and length_axon:
        content += f"4 2 -{radius} 0.0 0.0 {radius_axon} 1\n"
        content += f"5 2 -{radius + length_axon} 0.0 0.0 {radius_axon} 4\n"

    with open(str(morphology_path), "w+") as fp:
        fp.write(content)

    return morphology_path.resolve()



"""Helper functions used to find the path to the pdfs generated during the different steps
of the emodel pipeline"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import glob
import logging
import shutil
from pathlib import Path

logger = logging.getLogger(__name__)


def search_figure_path(pathname):
    """Search for a single pdf based on an expression"""

    matches = glob.glob(pathname)

    if not matches:
        logger.debug("No pdf for pathname %s", pathname)
        return None

    if len(matches) > 1:
        raise ValueError(f"More than one pdf for pathname {pathname}")

    return str(Path(matches[0]).resolve())


def search_figure_paths(pathname):
    """Search for at least one pdf based on an expression"""

    matches = glob.glob(pathname)

    if not matches:
        logger.debug("No pdf for pathname %s", pathname)
        return []

    return [str(Path(match).resolve()) for match in matches]


def figure_efeatures(emodel, protocol_name, efeature):
    """Get path to pdf representing the efeature extracted from ephys recordings"""

    path = f"./figures/{emodel}/efeatures_extraction/*{protocol_name}_{efeature}"

    pdf_amp = f"{path}_amp.pdf"
    pdf_amp_rel = f"{path}_amp_rel.pdf"

    return pdf_amp, pdf_amp_rel


def search_figure_efeatures(emodel, protocol_name, efeature):
    """Search for the pdf representing the efeature extracted from ephys recordings"""

    pdf_amp, pdf_amp_rel = figure_efeatures(emodel, protocol_name, efeature)

    pdf_amp = search_figure_path(pdf_amp)
    pdf_amp_rel = search_figure_path(pdf_amp_rel)

    return pdf_amp, pdf_amp_rel


def figure_emodel_optimisation(emodel_metadata, seed, use_allen_notation=True):
    """Get path for the pdf representing the convergence of the optimisation"""

    metadata_str = emodel_metadata.as_string(seed, use_allen_notation=use_allen_notation)
    fname = f"{metadata_str}__optimisation.pdf"

    return Path("./figures") / emodel_metadata.emodel / "optimisation" / fname


def search_figure_emodel_optimisation(emodel_metadata, seed, use_allen_notation=True):
    """Search for the pdf representing the convergence of the optimisation"""

    pathname = figure_emodel_optimisation(
        emodel_metadata, seed, use_allen_notation=use_allen_notation
    )

    return search_figure_path(str(pathname))


def figure_emodel_traces(emodel_metadata, seed, use_allen_notation=True):
    """Get path for the pdf representing the traces of an emodel"""

    metadata_str = emodel_metadata.as_string(seed, use_allen_notation=use_allen_notation)
    fname = f"{metadata_str}__traces.pdf"

    pathname = Path("./figures") / emodel_metadata.emodel / "traces" / "all" / fname
    pathname_val = Path("./figures") / emodel_metadata.emodel / "traces" / "validated" / fname

    return pathname, pathname_val


def search_figure_emodel_traces(emodel_metadata, seed, use_allen_notation=True):
    """Search for the pdf representing the traces of an emodel"""

    pathname, pathname_val = figure_emodel_traces(
        emodel_metadata, seed, use_allen_notation=use_allen_notation
    )

    return [search_figure_path(str(pathname)), search_figure_path(str(pathname_val))]


def figure_emodel_score(emodel_metadata, seed, use_allen_notation=True):
    """Get path for the pdf representing the scores of an emodel"""

    metadata_str = emodel_metadata.as_string(seed, use_allen_notation=use_allen_notation)
    fname = f"{metadata_str}__scores.pdf"

    pathname = Path("./figures") / emodel_metadata.emodel / "scores" / "all" / fname
    pathname_val = Path("./figures") / emodel_metadata.emodel / "scores" / "validated" / fname

    return pathname, pathname_val


def search_figure_emodel_score(emodel_metadata, seed, use_allen_notation=True):
    """Search for the pdf representing the scores of an emodel"""

    pathname, pathname_val = figure_emodel_score(
        emodel_metadata, seed, use_allen_notation=use_allen_notation
    )

    return [search_figure_path(str(pathname)), search_figure_path(str(pathname_val))]


def figure_emodel_thumbnail(emodel_metadata, seed, use_allen_notation=True):
    """Get path for the pdf representing the thumbnail of an emodel"""

    metadata_str = emodel_metadata.as_string(seed, use_allen_notation=use_allen_notation)
    fname = f"{metadata_str}__thumbnail.png"

    pathname = Path("./figures") / emodel_metadata.emodel / "thumbnail" / "all" / fname
    pathname_val = Path("./figures") / emodel_metadata.emodel / "thumbnail" / "validated" / fname

    return pathname, pathname_val


def search_figure_emodel_thumbnail(emodel_metadata, seed, use_allen_notation=True):
    """Search for the pdf representing the thumbnail of an emodel"""

    pathname, pathname_val = figure_emodel_thumbnail(
        emodel_metadata, seed, use_allen_notation=use_allen_notation
    )

    return [search_figure_path(str(pathname)), search_figure_path(str(pathname_val))]


def figure_emodel_bAP(emodel_metadata, seed, use_allen_notation=True):
    """Get path for the pdf representing the bAP of an emodel"""

    metadata_str = emodel_metadata.as_string(seed, use_allen_notation=use_allen_notation)
    fname = f"{metadata_str}__dendrite_backpropagation_fit_decay.pdf"

    pathname = Path("./figures") / emodel_metadata.emodel / "dendritic" / "all" / fname
    pathname_val = Path("./figures") / emodel_metadata.emodel / "dendritic" / "validated" / fname

    return pathname, pathname_val


def search_figure_emodel_bAP(emodel_metadata, seed, use_allen_notation=True):
    """Search for the pdf representing the bAP of an emodel"""

    pathname, pathname_val = figure_emodel_bAP(
        emodel_metadata, seed, use_allen_notation=use_allen_notation
    )

    return [search_figure_path(str(pathname)), search_figure_path(str(pathname_val))]


def figure_emodel_EPSP(emodel_metadata, seed, use_allen_notation=True):
    """Get path for the pdf representing the EPSP of an emodel"""

    metadata_str = emodel_metadata.as_string(seed, use_allen_notation=use_allen_notation)
    fname = f"{metadata_str}__dendrite_EPSP_attenuation_fit.pdf"

    pathname = Path("./figures") / emodel_metadata.emodel / "dendritic" / "all" / fname
    pathname_val = Path("./figures") / emodel_metadata.emodel / "dendritic" / "validated" / fname

    return pathname, pathname_val


def search_figure_emodel_EPSP(emodel_metadata, seed, use_allen_notation=True):
    """Search for the pdf representing the EPSP of an emodel"""

    pathname, pathname_val = figure_emodel_EPSP(
        emodel_metadata, seed, use_allen_notation=use_allen_notation
    )

    return [search_figure_path(str(pathname)), search_figure_path(str(pathname_val))]


def figure_emodel_ISI_CV(emodel_metadata, seed, use_allen_notation=True):
    """Get path for the pdf representing the ISI_CV fit of an emodel"""

    metadata_str = emodel_metadata.as_string(seed, use_allen_notation=use_allen_notation)
    fname = f"{metadata_str}__*ISI_CV_linear.pdf"

    pathname = Path("./figures") / emodel_metadata.emodel / "dendritic" / "all" / fname
    pathname_val = Path("./figures") / emodel_metadata.emodel / "dendritic" / "validated" / fname

    return pathname, pathname_val


def search_figure_emodel_ISI_CV(emodel_metadata, seed, use_allen_notation=True):
    """Search for the pdf representing the ISI_CV fit of an emodel"""

    pathname, pathname_val = figure_emodel_ISI_CV(
        emodel_metadata, seed, use_allen_notation=use_allen_notation
    )

    return list(search_figure_paths(str(pathname))) + list(search_figure_paths(str(pathname_val)))


def figure_emodel_rheobase(emodel_metadata, seed, use_allen_notation=True):
    """Get path for the pdf representing the rheobase fit of an emodel"""

    metadata_str = emodel_metadata.as_string(seed, use_allen_notation=use_allen_notation)
    fname = f"{metadata_str}__*bpo_threshold_current_linear.pdf"

    pathname = Path("./figures") / emodel_metadata.emodel / "dendritic" / "all" / fname
    pathname_val = Path("./figures") / emodel_metadata.emodel / "dendritic" / "validated" / fname

    return pathname, pathname_val


def search_figure_emodel_rheobase(emodel_metadata, seed, use_allen_notation=True):
    """Search for the pdf representing the rheobase fit of an emodel"""

    pathname, pathname_val = figure_emodel_rheobase(
        emodel_metadata, seed, use_allen_notation=use_allen_notation
    )

    return list(search_figure_paths(str(pathname))) + list(search_figure_paths(str(pathname_val)))


def figure_emodel_parameters(emodel_metadata, seed=None, use_allen_notation=True):
    """Get path for the pdf representing the distribution of the parameters of an emodel"""
    # pylint: disable=unused-argument

    metadata_str = emodel_metadata.as_string(use_allen_notation=use_allen_notation)
    fname = f"{metadata_str}__parameters_distribution.pdf"

    pathname = Path("./figures") / emodel_metadata.emodel / "distributions" / "all" / fname
    pathname_val = Path("./figures") / emodel_metadata.emodel / "distributions" / "validated"
    pathname_val = pathname_val / fname

    return pathname, pathname_val


def search_figure_emodel_parameters(emodel_metadata, use_allen_notation=True):
    """Search for the pdf representing the distribution of the parameters
    of an emodel"""

    pathname, pathname_val = figure_emodel_parameters(
        emodel_metadata, use_allen_notation=use_allen_notation
    )

    return [search_figure_path(str(pathname)), search_figure_path(str(pathname_val))]


def figure_emodel_parameters_evolution(emodel_metadata, seed=None, use_allen_notation=True):
    """Get path for the pdf representing the evolution of the parameters of an emodel"""

    metadata_str = emodel_metadata.as_string(seed, use_allen_notation=use_allen_notation)
    if seed is None:
        fname = f"{metadata_str}__all_seeds__evo_parameter_density.pdf"
    else:
        fname = f"{metadata_str}__evo_parameter_density.pdf"

    pathname = Path("./figures") / emodel_metadata.emodel / "parameter_evolution" / fname

    return pathname


def search_figure_emodel_parameters_evolution(emodel_metadata, seed=None, use_allen_notation=True):
    """Search for the pdf representing the evolution of the parameters of an emodel"""

    pathname = figure_emodel_parameters_evolution(
        emodel_metadata, seed, use_allen_notation=use_allen_notation
    )

    return search_figure_path(str(pathname))


def figure_emodel_currentscapes(emodel_metadata, seed, use_allen_notation=True):
    """Get path for the pdfs representing the currentscapes of an emodel"""

    metadata_str = emodel_metadata.as_string(seed, use_allen_notation=use_allen_notation)
    fname = f"{metadata_str}__currentscape*.pdf"

    pathname = Path("./figures") / emodel_metadata.emodel / "currentscape" / "all" / fname
    pathname_val = Path("./figures") / emodel_metadata.emodel / "currentscape" / "validated"
    pathname_val = pathname_val / fname

    return pathname, pathname_val


def search_figure_emodel_currentscapes(emodel_metadata, seed, use_allen_notation=True):
    """Search for the pdfs representing the currentscapes of an emodel"""

    pathname, pathname_val = figure_emodel_currentscapes(
        emodel_metadata, seed, use_allen_notation=use_allen_notation
    )

    return list(search_figure_paths(str(pathname))) + list(search_figure_paths(str(pathname_val)))


def copy_emodel_pdf_dependency_to_new_path(old_path, new_path, overwrite=False):
    """Copy a pdf dependency to new path using allen notation"""
    if old_path.is_file():
        if not new_path.is_file() or overwrite:
            new_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(old_path, new_path)


def copy_emodel_pdf_dependencies_to_new_path(
    old_metadata, new_metadata, old_allen_notation, new_allen_notation, seed, overwrite=False
):
    """Copy dependencies to new path using allen notation"""
    # pylint: disable=too-many-locals

    # do not have all and validated subfolders
    single_folder_fcts = [figure_emodel_optimisation, figure_emodel_parameters_evolution]
    # have all and validated subfolders
    two_folders_fcts = [
        figure_emodel_traces,
        figure_emodel_score,
        figure_emodel_parameters,
        figure_emodel_thumbnail,
    ]

    for fct in single_folder_fcts:
        old_path = fct(old_metadata, seed=seed, use_allen_notation=old_allen_notation)
        new_path = fct(new_metadata, seed=seed, use_allen_notation=new_allen_notation)
        copy_emodel_pdf_dependency_to_new_path(old_path, new_path, overwrite=overwrite)

    for fct in two_folders_fcts:
        old_path, old_path_val = fct(old_metadata, seed=seed, use_allen_notation=old_allen_notation)
        new_path, new_path_val = fct(new_metadata, seed=seed, use_allen_notation=new_allen_notation)
        copy_emodel_pdf_dependency_to_new_path(old_path, new_path, overwrite=overwrite)
        copy_emodel_pdf_dependency_to_new_path(old_path_val, new_path_val, overwrite=overwrite)

    # also check with seed = None for figure_emodel_parameters_evolution
    old_all_evo_path = figure_emodel_parameters_evolution(
        old_metadata, seed=None, use_allen_notation=old_allen_notation
    )
    new_all_evo_path = figure_emodel_parameters_evolution(
        new_metadata, seed=None, use_allen_notation=new_allen_notation
    )
    copy_emodel_pdf_dependency_to_new_path(old_all_evo_path, new_all_evo_path, overwrite=overwrite)

    # take into account that we have to search for currentscape plots
    # because we do not know a priori the protocols and locations
    old_currentscape_path = search_figure_emodel_currentscapes(
        old_metadata, seed, use_allen_notation=old_allen_notation
    )
    new_currentscape_path, new_currentscape_path_val = figure_emodel_currentscapes(
        new_metadata, seed, use_allen_notation=new_allen_notation
    )
    for old_path in old_currentscape_path:
        prot = str(Path(old_path).stem).rsplit("currentscape", maxsplit=1)[-1]
        if "/validated/" in str(old_path):
            new_path = str(new_currentscape_path_val).replace("*", prot)
        else:
            new_path = str(new_currentscape_path).replace("*", prot)
        copy_emodel_pdf_dependency_to_new_path(Path(old_path), Path(new_path), overwrite=overwrite)



"""Evaluation module

In addition, if the env variable USE_NEURODAMUS exists, it will load mechanisms
from the neurodamus modules, as done in BGLibPy. The functions below are adapted
from bglibpy.importer.py.
"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os


def check_hoc_lib():
    """Check of hoc library path is present"""

    if "HOC_LIBRARY_PATH" not in os.environ:
        raise RuntimeError(
            "BGLibPy: HOC_LIBRARY_PATH not found, this is required to find "
            "Neurodamus. Did you install neurodamus correctly ?"
        )
    return os.environ["HOC_LIBRARY_PATH"]


def import_mod_lib(_neuron):
    """Import mod files"""

    mod_lib_list = None
    if "BGLIBPY_MOD_LIBRARY_PATH" in os.environ:
        mod_lib_path = os.environ["BGLIBPY_MOD_LIBRARY_PATH"]
        mod_lib_list = mod_lib_path.split(":")
        for mod_lib in mod_lib_list:
            _neuron.h.nrn_load_dll(mod_lib)

    return mod_lib_list


def _nrn_disable_banner():
    """Disable Neuron banner"""

    import ctypes
    import importlib

    nrnpy_path = os.path.join(importlib.util.find_spec("neuron").submodule_search_locations[0])
    import glob

    hoc_so_list = glob.glob(os.path.join(nrnpy_path, "hoc*.so"))

    if len(hoc_so_list) != 1:
        raise FileNotFoundError(f"hoc shared library not found in {nrnpy_path}")

    hoc_so = hoc_so_list[0]
    nrndll = ctypes.cdll[hoc_so]
    ctypes.c_int.in_dll(nrndll, "nrn_nobanner_").value = 1


def import_neurodamus(_neuron):
    """Import neurodamus"""
    _neuron.h("objref simConfig")

    _neuron.h.load_file("stdrun.hoc")
    _neuron.h.load_file("defvar.hoc")
    _neuron.h.default_var("simulator", "NEURON")
    _neuron.h.load_file("Cell.hoc")
    _neuron.h.load_file("TDistFunc.hoc")
    _neuron.h.load_file("SerializedSections.hoc")
    _neuron.h.load_file("TStim.hoc")
    _neuron.h.load_file("ShowProgress.hoc")
    _neuron.h.load_file("SimSettings.hoc")
    _neuron.h.load_file("RNGSettings.hoc")

    _neuron.h("obfunc new_IClamp() { return new IClamp($1) }")
    _neuron.h("objref p")
    _neuron.h("p = new PythonObject()")

    _neuron.h("simConfig = new SimSettings()")


if os.getenv("USE_NEURODAMUS"):
    # this will loads mechanisms if proper neurodamus module is loaded
    check_hoc_lib()
    _nrn_disable_banner()
    import neuron

    import_mod_lib(neuron)
    import_neurodamus(neuron)



"""EModelMetadata class"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import string


def update_metadata_dict(metadata_dict, var, new_key):
    """Update metadata dict with new key and variable if not None."""
    if var is not None and var != "None":
        metadata_dict[new_key] = var


class EModelMetadata:
    """Contains the metadata of an emodel such as its e-model name or its brain region. These
    metadata can be understood as a unique identifier of an e-model.

    This class is a backend class and is not meant to be used directly by the user."""

    def __init__(
        self,
        emodel=None,
        etype=None,
        ttype=None,
        mtype=None,
        species=None,
        brain_region=None,
        iteration_tag=None,
        synapse_class=None,
        allen_notation=None,
    ):
        """Constructor

        Args:
            emodel (str): name of the e-model
            etype (str): name of the electric type of the e-model.
            ttype (str): name of the transcriptomic type of the e-model.
            mtype (str): name of the morphology type of the e-model.
            species (str): name of the species of the e-model.
            brain_region (str): name of the brain location of the e-model.
            iteration_tag (str): tag associated to the current run.
            synapse_class (str): synapse class (neurotransmitter)  of the e-model.
            allen_notation (str): Allen acronym for the brain region (Optional).
                Can be used to replace brain region in as_string().
        """

        if emodel is None and etype is None:
            raise ValueError("At least emodel or etype should be informed")

        self.emodel = emodel
        self.check_emodel_name()  # check if name complies with requirements
        self.etype = None if etype == "None" else etype
        self.ttype = None if ttype == "None" else ttype
        self.mtype = None if mtype == "None" else mtype
        self.species = None if species == "None" else species
        self.brain_region = None if brain_region == "None" else brain_region
        self.iteration = None if iteration_tag == "None" else iteration_tag
        self.synapse_class = None if synapse_class == "None" else synapse_class
        self.allen_notation = allen_notation

    def etype_annotation_dict(self):
        """Returns an etype annotation dict to be added to annotations list."""
        # TODO also add id from Cell Type Ontology in WebProtege (optional)
        return {
            "type": [
                "ETypeAnnotation",
                "Annotation",
            ],
            "hasBody": {
                "type": [
                    "EType",
                    "AnnotationBody",
                ],
                "label": self.etype,
            },
            "name": "E-type annotation",
        }

    def mtype_annotation_dict(self):
        """Returns an mtype annotation dict to be added to annotations list."""
        # TODO also add id from Cell Type Ontology in WebProtege (optional)
        return {
            "type": [
                "MTypeAnnotation",
                "Annotation",
            ],
            "hasBody": {
                "type": [
                    "MType",
                    "AnnotationBody",
                ],
                "label": self.mtype,
            },
            "name": "M-type annotation",
        }

    def ttype_annotation_dict(self):
        """Returns an ttype annotation dict to be added to annotations list."""
        # TODO also add id from Cell Type Ontology in WebProtege (optional)
        return {
            "type": [
                "TTypeAnnotation",
                "Annotation",
            ],
            "hasBody": {
                "type": [
                    "TType",
                    "AnnotationBody",
                ],
                "label": self.ttype,
            },
            "name": "T-type annotation",
        }

    def annotation_list(self):
        """Returns an annotation list containing mtype, etype and ttype annotations"""
        annotation_list = []
        if self.etype:
            annotation_list.append(self.etype_annotation_dict())
        if self.mtype:
            annotation_list.append(self.mtype_annotation_dict())
        if self.ttype:
            annotation_list.append(self.ttype_annotation_dict())

        return annotation_list

    def as_dict_for_resource(self):
        """Metadata as a dict, with keys consistent with nexus."""

        metadata_dict = {}

        update_metadata_dict(metadata_dict, self.emodel, "eModel")
        update_metadata_dict(metadata_dict, self.etype, "eType")
        update_metadata_dict(metadata_dict, self.ttype, "tType")
        update_metadata_dict(metadata_dict, self.mtype, "mType")
        update_metadata_dict(metadata_dict, self.iteration, "iteration")
        update_metadata_dict(metadata_dict, self.synapse_class, "synapseClass")
        # rename species into subject and brain_region into brainLocation
        update_metadata_dict(metadata_dict, self.species, "subject")
        update_metadata_dict(metadata_dict, self.brain_region, "brainLocation")
        # we do not want allen_notation in resource metadata

        return metadata_dict

    def as_dict_for_resource_legacy(self):
        """Metadata as a dict, with keys consistent with legacy nexus."""

        metadata_dict = {}

        for k, v in vars(self).items():
            # we do not want allen_notation in resource metadata
            if v and v != "None" and k != "allen_notation":
                # rename species into subject and brain_region into brainLocation
                if k == "species":
                    metadata_dict["subject"] = v
                elif k == "brain_region":
                    metadata_dict["brainLocation"] = v
                else:
                    metadata_dict[k] = v

        return metadata_dict

    def filters_for_resource(self):
        """Metadata used for filtering, without the annotation list"""
        return self.as_dict_for_resource()

    def filters_for_resource_legacy(self):
        """Legacy metadata used for filtering, without the annotation list"""
        return self.as_dict_for_resource_legacy()

    def for_resource(self):
        """Metadata to add to a resource to register.

        DO NOT use for filtering. For filtering, use self.filters_for_resource() instead.
        """

        metadata = self.as_dict_for_resource()

        metadata["annotation"] = self.annotation_list()

        return metadata

    def as_dict(self):
        """Metadata as dict."""
        return vars(self)

    def as_string(
        self, seed=None, use_allen_notation=True, replace_semicolons=True, replace_spaces=True
    ):
        s = ""

        for k in [
            "emodel",
            "etype",
            "ttype",
            "mtype",
            "species",
            "brain_region",
            "iteration",
        ]:
            v = getattr(self, k)
            if use_allen_notation and k == "brain_region" and self.allen_notation:
                v = self.allen_notation
            if v:
                if isinstance(v, int):
                    v = str(v)
                s += f"{k}={v.replace('/', '')}__"

        if seed not in [None, "None"]:
            s += f"seed={seed}__"

        # can have ':' in mtype. Replace this character.
        if replace_semicolons:
            s = s.replace(":", "_")

        # also replace spaces if any
        if replace_spaces:
            s = s.replace(" ", "_")

        return s[:-2]

    def check_emodel_name(self):
        """Check if name complies with requirements:
        https://nrn.readthedocs.io/en/8.2.3/guide/hoc_chapter_11_old_reference.html#names
        """

        allowed_chars = string.ascii_letters + string.digits + "_"
        translate_args = str.maketrans("", "", allowed_chars)

        if (
            self.emodel == ""
            or self.emodel[0] not in string.ascii_letters
            or not str(self.emodel).translate(translate_args) == ""
        ):
            raise TypeError(
                f"Emodel: name {self.emodel} provided to constructor does not comply "
                "with the rules for Neuron template name: name should be "
                "alphanumeric "
                "non-empty string, underscores are allowed, "
                "first char should be letter"
            )



"""Noise stimulus class"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from bluepyemodel.ecode.stimulus import BPEM_stimulus

logger = logging.getLogger(__name__)


class NoiseMixin(BPEM_stimulus):
    """Noise current stimulus"""

    name = "Noise"

    def __init__(self, location):
        """Constructor

        Args:
            location(Location): location of stimulus
        """
        super().__init__(
            location=location,
        )

    @property
    def total_duration(self):
        return self.time_series[-1]

    @property
    def stim_start(self):
        return 0.0

    @property
    def stim_end(self):
        return self.time_series[-1]

    def generate(self, dt=0.1):
        """Return current time series"""
        holding_current = self.holding_current if self.holding_current is not None else 0

        if dt != 0.1:
            raise ValueError(f"For eCode {self.name}, dt has to be 0.1ms.")

        current = holding_current + self.current_series * (self.mu / 2.0) + self.mu

        return self.time_series, current



"""Morphology utils."""

"""
Copyright 2023-2024, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


def get_apical_point_soma_distance(morph_path):
    """Get the euclidian distance between soma and apical point given a morphology.

    Args:
        morph_path (str): path to the morphology

    Returns:
        float: euclidian distance between soma and apical point
    """
    import morph_tool
    import morphio
    import neurom

    morphio_morph = morphio.Morphology(morph_path)
    apical_point = morph_tool.apical_point.apical_point_position(morphio_morph)
    return neurom.morphmath.point_dist(apical_point, morphio_morph.soma.center)


def get_apical_max_radial_distance(morph_path):
    """Returns the max radial distance of the apical dendrites."""
    import neurom

    neurom_morph = neurom.load_morphology(morph_path)
    return neurom.features.get(
        "max_radial_distance", neurom_morph, neurite_type=neurom.APICAL_DENDRITE
    )


def get_basal_and_apical_max_radial_distances(morph_path):
    """Returns the max radial distances of the apical and basal dendrites."""
    import neurom

    neurom_morph = neurom.load_morphology(morph_path)
    basal_length = neurom.features.get(
        "max_radial_distance", neurom_morph, neurite_type=neurom.BASAL_DENDRITE
    )
    apical_length = neurom.features.get(
        "max_radial_distance", neurom_morph, neurite_type=neurom.APICAL_DENDRITE
    )
    return basal_length, apical_length


def get_hotspot_location(morph_path, hotspot_percent=20.0):
    """Get hot spot begin and end in terms of soma distance.

    Calcium hot spot should be in distal apical trunk, and in primary and secondary tufts,
    i.e. around apical point, according to Larkum and Zhu, 2002.

    Attention! Apical point detection is not always accurate.

    Args:
        morph_path (str): path to the morphology
        hotspot_percent (float): percentage of the radial apical distance that is in the hot spot.
            Here, we assume that the hotspot size is dependent on the apical radial distance.
            20% is in accordance with experiments from Larkum and Zhu, 2002
    """
    ap_soma_dist = get_apical_point_soma_distance(morph_path)
    apical_length = get_apical_max_radial_distance(morph_path)
    hotspot_halfsize = apical_length * hotspot_percent / 2.0 / 100.0

    return max((0.0, ap_soma_dist - hotspot_halfsize)), ap_soma_dist + hotspot_halfsize



"""MultiProtocol eFeature Utils"""

"""
Copyright 2023-2024, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging

logger = logging.getLogger("__main__")


def get_distances_from_recording_name(rec_name):
    """Get apical distances from recording name."""
    isolate_list = rec_name.split("[")[1].split("]")[0]
    int_list = [int(num) for num in isolate_list.split(",")]
    int_list.insert(0, 0)
    return int_list


def get_soma_protocol_from_protocol_name(prot_name, amplitude):
    """Get recording name at soma from multi-locations protocols' name

    Args:
        prot_name (str): protocol name
        amplitude (int): amplitude
    """
    from bluepyemodel.ecode import eCodes

    # find ecode in protocol name
    for ecode in eCodes:
        idx = prot_name.lower().find(ecode)
        if idx > -1:
            # get ecode with capitalized letters
            prot_name = prot_name[idx : idx + len(ecode)]
            return f"{prot_name}_{amplitude}"

    raise ValueError(f"Could not find ecode in {prot_name}")


def get_soma_protocol_from_recording_name(rec_name, amplitude):
    """Get recording name at soma from multi-locations protocols' name"""
    return f"{get_soma_protocol_from_protocol_name(rec_name.split('.')[0], amplitude)}.soma.v"


def split_protocol_name_with_location_list(prot_name):
    """Split full protocol name into 'real' protocol name, location list and amplitude.

    Args:
        prot_name (str): full protocol name containing location list
            e.g. LocalInjectionIDrestapic[050,080,110,200,340]_100
    """
    prot_name, tmp = prot_name.split("[")
    isolate_list, amp = tmp.split("]")
    amplitude = int(amp.split("_")[1])
    return prot_name, isolate_list, amplitude


def get_protocol_list_from_protocol_name(prot_name):
    """Reconstruct protocol list from protocol name"""
    prot_name, isolate_list, amplitude = split_protocol_name_with_location_list(prot_name)
    prot_list = [f"{prot_name}{distance}_{amplitude}" for distance in isolate_list.split(",")]

    prot_list.insert(0, get_soma_protocol_from_protocol_name(prot_name, amplitude))

    return prot_list


def get_protocol_list_from_recording_name(rec_name):
    """Reconstruct recording list from recording name."""
    prot, loc, var = rec_name.split(".")
    prot_name, isolate_list, amplitude = split_protocol_name_with_location_list(prot)
    loc = loc.split("[")[0]
    prot_list = [
        (
            f"{prot_name}{distance}_{amplitude}.{loc}.{var}"
            if loc == "soma"
            else f"{prot_name}{distance}_{amplitude}.{loc}{distance}.{var}"
        )
        for distance in isolate_list.split(",")
    ]

    prot_list.insert(0, get_soma_protocol_from_recording_name(rec_name, amplitude))

    return prot_list


def get_locations_from_recording_name(rec_name):
    prot, loc, _ = rec_name.split(".")
    _, isolate_list, _ = split_protocol_name_with_location_list(prot)
    loc = loc.split("[")[0]
    loc_list = [f"{loc}{distance}" for distance in isolate_list.split(",")]
    loc_list.insert(0, "soma")
    return loc_list



"""eCode init script"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .apwaveform import APWaveform
from .comb import Comb
from .customfromfile import CustomFromFile
from .dehyperpol import DeHyperpol
from .dendrite import BAC
from .dendrite import DendriticStep
from .dendrite import Synaptic
from .firepattern import FirePattern
from .hyperdepol import HyperDepol
from .idrest import IDrest
from .iv import IV
from .negcheops import NegCheops
from .noiseou3 import NoiseOU3
from .poscheops import PosCheops
from .probampanmda_ems import ProbAMPANMDA_EMS
from .ramp import Ramp
from .random_square_inputs import MultipleRandomStepInputs
from .sahp import sAHP
from .sinespec import SineSpec
from .spikerec import SpikeRecMultiSpikes
from .square import BPOSquarePulse
from .subwhitenoise import SubWhiteNoise
from .thresholdaddition import ThresholdAddition
from .whitenoise import WhiteNoise

# The ecode names have to be lower case only, to avoid having to
# define duplicates.
eCodes = {
    "spontaneous": IDrest,
    "step": IDrest,
    "idrest": IDrest,
    "idthres": IDrest,
    "idthresh": IDrest,
    "spontaps": IDrest,
    "sponnohold30": IDrest,
    "sponhold30": IDrest,
    "rinholdcurrent": IDrest,
    "bap": IDrest,
    "spikerecmultispikes": SpikeRecMultiSpikes,
    "iv": IV,
    "spikerec": IDrest,
    "apwaveform": APWaveform,
    "firepattern": FirePattern,
    "sahp": sAHP,
    "idhyperpol": sAHP,
    "irdepol": sAHP,
    "irhyperpol": sAHP,
    "iddepol": sAHP,
    "hyperdepol": HyperDepol,
    "dehyperpol": DeHyperpol,
    "poscheops": PosCheops,
    "negcheops": NegCheops,
    "ramp": Ramp,
    "ap_thresh": Ramp,
    "apthresh": Ramp,
    "sinespec": SineSpec,
    "subwhitenoise": SubWhiteNoise,
    "noiseou3": NoiseOU3,
    "whitenoise": WhiteNoise,
    "highfreq": Comb,
    "synaptic": Synaptic,
    "bac": BAC,
    "dendritic": DendriticStep,
    "randomsteps": MultipleRandomStepInputs,
    "custom": CustomFromFile,
    "startnohold": IDrest,
    "thresholdaddition": ThresholdAddition,
    "probampanmda_ems": ProbAMPANMDA_EMS,
    "bposquarepulse": BPOSquarePulse,
}

fixed_timestep_eCodes = ["probampanmda_ems"]



"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import argparse
import glob

from bluepyemodel.export_emodel.export_emodel import export_emodels_hoc
from bluepyemodel.export_emodel.export_emodel import export_emodels_sonata
from bluepyemodel.emodel_pipeline.emodel_pipeline import EModel_pipeline
from bluepyemodel.efeatures_extraction.targets_configurator import TargetsConfigurator
from targets import file_type, filenames, ecodes_metadata, targets, protocols_rheobase

logger = logging.getLogger()


def configure_targets(access_point):
    files_metadata = []

    if file_type == "ibw":
        fs = glob.glob(f"{filenames[0]}/*ch0*.ibw")
        for filename in fs:
            fn = filename.split("/")[-1]
            for ecode in ecodes_metadata:
                if ecode in fn:
                    files_metadata.append(
                        {
                            "cell_name": "cell1",
                            "filename": filename.split("/")[-1].split(".")[0],
                            "ecodes": {ecode: ecodes_metadata[ecode]},
                            "other_metadata": {
                                "v_file": filename,
                                # IMPORTANT: modify "ch1" with your number of channel in your ibw files
                                "i_file": filename.replace("ch0", "ch1"),
                                "i_unit": "A",
                                "v_unit": "V",
                                "t_unit": "s",
                            },
                        }
                    )
    elif file_type == "nwb":
        files_metadata = []
        for filename in filenames:
            files_metadata.append(
                {
                    "cell_name": filename.split("/")[-1].split(".")[0],
                    "filepath": filename,
                    "ecodes": ecodes_metadata,
                }
            )
    else:
        raise ValueError(
            f"file type {file_type} is not supported in this current pipeline: "
            "Expected 'ibw' or 'nwb'. "
            "For other types of format please modify configure_targets "
            "in pipeline.py."
        )

    targets_formated = []
    for ecode in targets:
        for amplitude in targets[ecode]["amplitudes"]:
            for efeature in targets[ecode]["efeatures"]:
                protocol_name = ecode
                # remove ohmic_input_resistance_vb_ssse for IV_0 protocol (RMPProtocol). we only need voltage_base
                if not (
                    protocol_name == "IV"
                    and amplitude == 0
                    and efeature == "ohmic_input_resistance_vb_ssse"
                ):
                    targets_formated.append(
                        {
                            "efeature": efeature,
                            "protocol": protocol_name,
                            "amplitude": amplitude,
                            "tolerance": 20,
                        }
                    )

    if not files_metadata:
        raise FileNotFoundError(
            "Cannot find electrophysiological experimental data. "
            "Please provide them, with the correct path to them in targets.py."
        )

    configurator = TargetsConfigurator(access_point)
    configurator.new_configuration(files_metadata, targets_formated, protocols_rheobase)
    configurator.save_configuration()


def get_parser():

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="Pipeline for emodel creation",
    )

    parser.add_argument(
        "--step",
        type=str,
        required=True,
        choices=["extract", "optimise", "analyse", "export_hoc", "export_sonata"],
    )
    parser.add_argument("--emodel", type=str, required=True)
    parser.add_argument("--etype", type=str, required=False, default=None)
    parser.add_argument("--mtype", type=str, required=False, default=None)
    parser.add_argument("--ttype", type=str, required=False, default=None)
    parser.add_argument("--seed", type=int, default=1)
    parser.add_argument("--githash", type=str, required=False, default=None)
    parser.add_argument("--use_ipyparallel", action="store_true", default=False)
    parser.add_argument("--use_multiprocessing", action="store_true", default=False)
    parser.add_argument("-v", "--verbose", action="count", dest="verbosity", default=0)

    return parser


def main():

    args = get_parser().parse_args()

    species = "rat"
    brain_region = "SSCX"

    recipes_path = "./config/recipes.json"

    logging.basicConfig(
        level=(logging.WARNING, logging.INFO, logging.DEBUG)[args.verbosity],
        handlers=[logging.StreamHandler()],
    )
    logger.setLevel((logging.WARNING, logging.INFO, logging.DEBUG)[args.verbosity])

    pipeline = EModel_pipeline(
        emodel=args.emodel,
        etype=args.etype,
        mtype=args.mtype,
        ttype=args.ttype,
        species=species,
        brain_region=brain_region,
        recipes_path=recipes_path,
        iteration_tag=args.githash,
        use_ipyparallel=args.use_ipyparallel,
        use_multiprocessing=args.use_multiprocessing,
    )

    if args.step == "extract":
        configure_targets(pipeline.access_point)
        pipeline.extract_efeatures()

    elif args.step == "optimise":
        pipeline.optimise(seed=args.seed)

    elif args.step == "analyse":
        pipeline.store_optimisation_results()
        pipeline.validation()
        pipeline.plot(only_validated=False)

    elif args.step == "export_hoc":
        export_emodels_hoc(
            pipeline.access_point,
            only_validated=False,
            only_best=False,
            seeds=[args.seed],
        )

    elif args.step == "export_sonata":
        export_emodels_sonata(
            pipeline.access_point,
            only_validated=False,
            only_best=False,
            seeds=[args.seed],
            map_function=pipeline.mapper,
        )


if __name__ == "__main__":
    main()



"""Luigi config classes."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import luigi


class EmodelAPIConfig(luigi.Config):
    """Configuration of emodel api database."""

    api = luigi.Parameter(default="local")

    # local parameters
    emodel_dir = luigi.OptionalParameter(default="./")
    recipes_path = luigi.OptionalParameter(default=None)
    final_path = luigi.OptionalParameter(default="./final.json")
    legacy_dir_structure = luigi.BoolParameter(default=False)

    # nexus parameters
    forge_path = luigi.OptionalParameter(default=None)
    forge_ontology_path = luigi.OptionalParameter(default=None)
    nexus_project = luigi.OptionalParameter(default="emodel_pipeline")
    nexus_organisation = luigi.OptionalParameter(default="demo")
    nexus_endpoint = luigi.OptionalParameter(default="https://bbp.epfl.ch/nexus/v1")

    def __init__(self, *args, **kwargs):
        """Init."""
        super().__init__(*args, **kwargs)

        if self.api == "local":
            if self.recipes_path is None:
                raise ValueError("recipes_path cannot be None when api is set to 'local'")

            self.api_args = {
                "emodel_dir": self.emodel_dir,
                "recipes_path": self.recipes_path,
                "final_path": self.final_path,
                "legacy_dir_structure": self.legacy_dir_structure,
            }

        if self.api == "nexus":
            if self.forge_path is None:
                raise ValueError("forge_path cannot be None when api is set to 'nexus'")
            if self.forge_ontology_path is None:
                raise ValueError("forge_ontology_path cannot be None when api is set to 'nexus'")

            self.api_args = {
                "forge_path": self.forge_path,
                "forge_ontology_path": self.forge_ontology_path,
                "project": self.nexus_project,
                "organisation": self.nexus_organisation,
                "endpoint": self.nexus_endpoint,
            }



"""Utils"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import pickle
from pathlib import Path

import numpy

from bluepyemodel.ecode import IDrest
from bluepyemodel.ecode import eCodes

logger = logging.getLogger("__main__")


def checkpoint_path_exists(checkpoint_path):
    """Returns True if checkpoint path exists, False if not.

    Args:
        checkpoint_path (str or Path): checkpoint path
    """
    checkpoint_path = Path(checkpoint_path)
    return (
        checkpoint_path.is_file()
        or checkpoint_path.with_suffix(checkpoint_path.suffix + ".tmp").is_file()
    )


def get_checkpoint_path(metadata, seed=None):
    """Get checkpoint path. Use legacy format if any is found, else use latest format."""
    base_path = f"./checkpoints/{metadata.emodel}/{metadata.iteration}/"
    # legacy case 1 (2023.05.11 - 2023.10.19)
    filename = metadata.as_string(
        seed=seed, use_allen_notation=False, replace_semicolons=False, replace_spaces=False
    )
    full_path = f"{base_path}{filename}.pkl"

    # legacy case 0 (before 2023.05.11)
    if checkpoint_path_exists(get_legacy_checkpoint_path(full_path)):
        full_path = get_legacy_checkpoint_path(full_path)

    # legacy case 2 (2023.10.19 - 2024.02.14)
    if not checkpoint_path_exists(full_path):
        filename = metadata.as_string(
            seed=seed, use_allen_notation=True, replace_semicolons=False, replace_spaces=False
        )
        full_path = f"{base_path}{filename}.pkl"

    # legacy case 3 (2024.02.14 - 2024.05.29)
    if not checkpoint_path_exists(full_path):
        filename = metadata.as_string(
            seed=seed, use_allen_notation=True, replace_semicolons=True, replace_spaces=False
        )
        full_path = f"{base_path}{filename}.pkl"

    # Up-to-date checkpoint path (after 2024.05.29)
    if not checkpoint_path_exists(full_path):
        filename = metadata.as_string(
            seed=seed, use_allen_notation=True, replace_semicolons=True, replace_spaces=True
        )
        full_path = f"{base_path}{filename}.pkl"

    return full_path


def get_legacy_checkpoint_path(checkpoint_path):
    """Get legacy checkpoint path from checkpoint path"""

    filename = Path(checkpoint_path).stem
    return f"./checkpoints/{filename}.pkl"


def make_dir(path_dir):
    """Creates directory if it does not exist"""
    p = Path(path_dir)
    if not (p.is_dir()):
        logger.info("Creating directory %s.", p)
        p.mkdir(parents=True, exist_ok=True)


def yesno(question):
    """Ask a Yes/No question"""

    prompt = f"{question} ? (y/n): "
    ans = input(prompt).strip().lower()

    if ans not in ["y", "n"]:
        print(f"{ans} is invalid, please try again...")
        return yesno(question)

    if ans == "y":
        return True

    return False


def get_seed_from_checkpoint_path(path):
    """Get seed from checkpoint path name. default seed is 0 if not found."""

    if path.endswith(".tmp"):
        path = path.replace(".tmp", "")

    filename = Path(path).stem.split("__")

    search_str = "seed="
    seed = next((e.replace(search_str, "") for e in filename if search_str in e), 0)

    return int(seed)


def read_checkpoint(checkpoint_path):
    """Reads a BluePyOpt checkpoint file"""

    p = Path(checkpoint_path)
    p_tmp = p.with_suffix(p.suffix + ".tmp")

    try:
        with open(str(p), "rb") as checkpoint_file:
            run = pickle.load(checkpoint_file, encoding="latin1")
            seed = get_seed_from_checkpoint_path(str(p))
    except EOFError:
        try:
            with open(str(p_tmp), "rb") as checkpoint_tmp_file:
                run = pickle.load(checkpoint_tmp_file, encoding="latin1")
                seed = get_seed_from_checkpoint_path(str(p_tmp))
        except EOFError:
            logger.error(
                "Cannot store model. Checkpoint file %s does not exist or is corrupted.",
                checkpoint_path,
            )

    return run, seed


def format_protocol_name_to_list(protocol_name):
    """Make sure that the name of a protocol is a list [protocol_name, amplitude]"""

    if isinstance(protocol_name, str):
        try:
            name_parts = protocol_name.split("_")
            if name_parts[-1] == "hyp":
                amplitude = float(name_parts[-2])
                name_parts.pop(-2)
                name = "_".join(name_parts)
            else:
                name = "_".join(name_parts[:-1])
                amplitude = float(protocol_name.split("_")[-1])
        except ValueError:
            return protocol_name, None
        return name, amplitude

    if isinstance(protocol_name, list):
        return protocol_name

    raise TypeError("protocol_name should be a string or a list.")


def are_same_protocol(name_a, name_b):
    """Check if two protocol names or list are equal. Eg: is IV_0.0 the same as IV_0 and
    the same as ["IV", 0.0]."""

    if name_a is None or name_b is None:
        return False

    amps = []
    ecodes = []

    for name in [name_a, name_b]:
        tmp_p = format_protocol_name_to_list(name)
        ecodes.append(tmp_p[0])
        amps.append(tmp_p[1])

    if ecodes[0] == ecodes[1] and numpy.isclose(amps[0], amps[1]):
        return True
    return False


def get_mapped_protocol_name(protocol_name, protocols_mapping):
    """Returns the mapped protocol name from protocols_mapping if available;
    otherwise, returns the original name."""

    if protocols_mapping:
        for p in protocols_mapping:
            if protocol_name.lower() in p.lower():
                return protocols_mapping[p]
    return protocol_name


def select_rec_for_thumbnail(rec_names, additional_step_prots=None, thumbnail_rec=None):
    """Select a recording for thumbnail.

    Select the step protocol with lowest positive amplitude, so that delay is visible if present.

    Args:
        rec_names (set): the names of the recordings, following this naming convention:
            protocol_name_amplitude.location.variable
        additional_step_prots (list): step protocol names to look for (other than defaults ones)
        thumbnail_rec (str): recording name to use for thumbnail if present
    """
    if thumbnail_rec is not None:
        if thumbnail_rec in rec_names:
            return thumbnail_rec
        logger.warning(
            "Could not find %s in recording names. Will use another recording for thumbnail plot.",
            thumbnail_rec,
        )
    selected_rec = ""
    selected_amp = numpy.inf
    step_prots = [prot_name for prot_name, prot in eCodes.items() if prot is IDrest]
    if additional_step_prots:
        step_prots = step_prots + additional_step_prots

    for rec_name in rec_names:
        # TODO: have a more proper way to remove non somatic injections
        if "LocalInjection" not in rec_name and any(
            step_prot.lower() in rec_name.lower() for step_prot in step_prots
        ):
            prot_name = rec_name.split(".")[0]
            try:
                _, rec_amp = format_protocol_name_to_list(prot_name)
                if 0 < rec_amp < selected_amp:
                    selected_rec = rec_name
                    selected_amp = rec_amp
            except (TypeError, ValueError):
                logger.warning("Could not find amplitude in %s, skipping it.", prot_name)

    if selected_rec == "":
        if len(rec_names) < 1:
            raise ValueError("No recording in recording_names. Can not plot thumbnail.")
        logger.warning("Could not find any step protocol in recording. Will take the first one.")
        return next(iter(rec_names))

    logger.debug("Selected %s for thumbnail", selected_rec)

    return selected_rec



"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# Specify the file type being used;
# the current pipeline.py only supports "ibw" or "nwb" files.
# modify the pipeline.py file if you want to use other file types
file_type = "ibw"  # Change this to "nwb" if using NWB files

filenames = [
    # "./ephys_data/YOUR_NWB_FILE.nwb"
    "./ephys_data/C060109A1-SR-C1/"  # if using ibw files, please provide the path to the folder containing the .ibw files
]

# FEATURE TIMINGS IN FEATURE EXTRACTION
# ----------------------------------
# Note regarding feature timings (ton, tmid, tmid2, toff):
# While these timings are not mandatory parameters, they significantly enhance the reliability of feature extraction.
# Bluepyefe includes an algorithm for their automatic detection; however, this can occasionally fail, especially with noisy data.
# Therefore, if known, providing these timings is recommended for increased accuracy and reliability of the results.

# LIQUID JUNCTION POTENTIAL (LJP) CORRECTION
# ----------------------------------
# Liquid Junction Potential (ljp) is a correction factor that is applied to the voltage measurements in experimental data.
# It accounts for the potential difference caused by the liquid junction in electrophysiological recordings.
# It is crucial to provide this correction value if you are aware that your data requires an LJP adjustment.
# Failure to apply a known LJP correction can result in misinterpretation of the data due to inaccurate voltage readings.
ecodes_metadata = {
    "IDthresh": {"ljp": 14.0, "ton": 700, "toff": 2700},
    "IDrest": {"ljp": 14.0, "ton": 700, "toff": 2700},
    "IV": {"ljp": 14.0, "ton": 20, "toff": 1020},
    "APWaveform": {"ljp": 14.0, "ton": 5, "toff": 55},
    "sAHP": {"ljp": 14.0, "ton": 25, "tmid": 520, "tmid2": 720, "toff": 2720},
}


protocols_rheobase = ["IDthresh"]

targets = {
    "IDrest": {
        "amplitudes": [150, 250],
        "efeatures": [
            "Spikecount",
            "mean_frequency",
            "time_to_first_spike",
            "time_to_last_spike",
            "inv_time_to_first_spike",
            "inv_first_ISI",
            "inv_second_ISI",
            "inv_third_ISI",
            "inv_last_ISI",
            "AHP_depth",
            "AHP_time_from_peak",
            "min_AHP_values",
            "depol_block_bool",
            "voltage_base",
        ],
    },
    "IV": {
        "amplitudes": [0, -20, -100],
        "efeatures": [
            "voltage_base",
            "ohmic_input_resistance_vb_ssse",
        ],
    },
    "APWaveform": {
        "amplitudes": [280],
        "efeatures": ["AP_amplitude", "AP1_amp", "AP_duration_half_width", "AHP_depth"],
    },
    "sAHP": {
        "amplitudes": [220],
        "efeatures": ["mean_frequency", "voltage_base", "depol_block_bool"],
    },
}



"""EModelScript class"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class EModelScript:
    """Contains an emodel hoc file path."""

    def __init__(self, hoc_file_path=None, seed=None, workflow_id=None):
        """Init

        Args:
            hoc_file_path (str): path to the hoc file of the emodel
            seed (str): seed used during optimisation for this emodel.
            workflow_id (str): id of the emodel workflow resource
        """
        self.hoc_file_path = hoc_file_path
        self.seed = seed
        self.workflow_id = workflow_id

    def get_related_nexus_ids(self):
        return {
            "generation": {
                "type": "Generation",
                "activity": {
                    "type": "Activity",
                    "followedWorkflow": {
                        "type": "EModelWorkflow",
                        "id": self.workflow_id,
                    },
                },
            }
        }

    def as_dict(self):
        return {
            "nexus_distributions": [self.hoc_file_path],
            "seed": self.seed,
        }



"""EFeatureConfiguration"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class EFeatureConfiguration:
    """Container for the definition of an EFeature"""

    def __init__(
        self,
        efel_feature_name,
        protocol_name,
        recording_name,
        mean,
        efeature_name=None,
        efel_settings=None,
        threshold_efeature_std=None,
        original_std=None,
        std=None,
        sample_size=None,
        default_std_value=1e-3,
        weight=1.0,
    ):
        """Init.

        The arguments efeatures and protocols are expected to be in the format used for the
        storage of the fitness calculator configuration. To store the results of an extraction,
        use the method init_from_bluepyefe.

        Args:
            efel_feature_name (str): name of the eFEl feature.
            protocol_name (str): name of the protocol to which the efeature is associated. For
                example "Step_200".
            recording_name (str or dict): name of the recording(s) of the protocol. For
                example: "soma.v" or if and only if the feature depends on several recordings:
                {"": "soma.v", "location_AIS": "axon.v"}.
            mean (float): mean of the efeature.
            original_std (float): unmodified standard deviation of the efeature
            std (float): kept for legacy purposes.
            efeature_name (str):given name for this specific feature. Can be different
                from the efel efeature name.
            efel_settings (dict): eFEl settings.
            threshold_efeature_std (float): lower limit for the std expressed as a percentage of
                the mean of the features value (optional).
            sample_size (float): number of data point that were used to compute the present
                average and standard deviation.
            weight (float): weight of the efeature.
                Basically multiplies the score of the efeature by this value.
        """

        self.efel_feature_name = efel_feature_name
        self.protocol_name = protocol_name
        self.recording_name = recording_name
        self.threshold_efeature_std = threshold_efeature_std
        self.default_std_value = default_std_value

        self.mean = mean
        self.original_std = original_std if original_std is not None else std
        self.sample_size = sample_size

        self.efeature_name = efeature_name
        self.weight = weight

        if efel_settings is None:
            self.efel_settings = {"strict_stiminterval": True}
        else:
            self.efel_settings = efel_settings

    @property
    def name(self):
        n = self.efeature_name if self.efeature_name else self.efel_feature_name
        if isinstance(self.recording_name, dict):
            return f"{self.protocol_name}.{self.recording_name['']}.{n}"
        return f"{self.protocol_name}.{self.recording_name}.{n}"

    @property
    def recording_name_for_instantiation(self):
        if isinstance(self.recording_name, dict):
            return {k: f"{self.protocol_name}.{v}" for k, v in self.recording_name.items()}
        return {"": f"{self.protocol_name}.{self.recording_name}"}

    @property
    def std(self):
        """Limit the standard deviation with a lower bound equal to a percentage of the mean."""

        if self.threshold_efeature_std is None:
            return self.original_std

        if self.mean == 0.0:
            if self.threshold_efeature_std:
                return self.threshold_efeature_std
            return self.default_std_value

        limit = abs(self.threshold_efeature_std * self.mean)
        if self.original_std < limit:
            return limit

        return self.original_std

    def as_dict(self):
        """Dictionary form"""

        return vars(self)



"""BPEM_stimulus class"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from bluepyopt.ephys.stimuli import Stimulus

logger = logging.getLogger(__name__)


class BPEM_stimulus(Stimulus):
    """Abstract current stimulus"""

    name = ""

    def __init__(self, location):
        """Constructor
        Args:
            total_duration (float): total duration of the stimulus in ms
            location(Location): location of stimulus
        """

        super().__init__()

        self.location = location

        self.iclamp = None
        self.current_vec = None
        self.time_vec = None

    @property
    def stim_start(self):
        return 0.0

    @property
    def stim_end(self):
        return 0.0

    @property
    def amplitude(self):
        return 0.0

    def instantiate(self, sim=None, icell=None):
        """Run stimulus"""

        time_series, current_series = self.generate(dt=0.1)

        icomp = self.location.instantiate(sim=sim, icell=icell)

        self.iclamp = sim.neuron.h.IClamp(icomp.x, sec=icomp.sec)
        self.iclamp.dur = time_series[-1]

        self.current_vec = sim.neuron.h.Vector()
        self.time_vec = sim.neuron.h.Vector()

        for t, i in zip(time_series, current_series):
            self.time_vec.append(t)
            self.current_vec.append(i)

        self.iclamp.delay = 0
        self.current_vec.play(
            self.iclamp._ref_amp,  # pylint:disable=W0212
            self.time_vec,
            1,
            sec=icomp.sec,
        )

    def destroy(self, sim=None):  # pylint:disable=W0613
        """Destroy stimulus"""
        self.iclamp = None
        self.time_vec = None
        self.current_vec = None

    def generate(self, dt=0.1):  # pylint:disable=W0613
        """Return current time series"""
        return [], []

    def __str__(self):
        """String representation"""
        return f"{self.name} current played at {self.location}"



"""Targets Configurator"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from bluepyemodel.access_point.local import LocalAccessPoint
from bluepyemodel.efeatures_extraction.auto_targets import get_auto_target_from_presets
from bluepyemodel.efeatures_extraction.targets_configuration import TargetsConfiguration
from bluepyemodel.tools.utils import yesno

logger = logging.getLogger(__name__)


class TargetsConfigurator:
    """Handles the loading, saving and modification of a targets configuration"""

    def __init__(self, access_point, configuration=None):
        """Creates a targets configuration, which includes the ephys files and targets

        Args:
            access_point (DataAccessPoint): access point to the emodel data.
            configuration (TargetsConfiguration): a pre-existing configuration.
        """

        self.access_point = access_point
        self.configuration = configuration

    def new_configuration(
        self,
        files=None,
        targets=None,
        protocols_rheobase=None,
        auto_targets=None,
        protocols_mapping=None,
    ):
        """Create a new configuration"""

        if self.configuration is not None:
            self.delete_configuration()

        available_traces = self.access_point.get_available_traces()
        available_efeatures = self.access_point.get_available_efeatures()

        self.configuration = TargetsConfiguration(
            files=files,
            targets=targets,
            protocols_rheobase=protocols_rheobase,
            available_traces=available_traces,
            available_efeatures=available_efeatures,
            auto_targets=auto_targets,
            protocols_mapping=protocols_mapping,
        )

    def load_configuration(self):
        """Load a previously registered configuration"""

        if isinstance(self.access_point, LocalAccessPoint):
            raise NotImplementedError(
                "Loading configuration is not yet implemented for local access point"
            )

        self.access_point.get_targets_configuration()

    def save_configuration(self):
        """Save the configuration. The saving medium depends of the access point."""

        if self.configuration:
            if self.configuration.is_configuration_valid:
                self.access_point.store_targets_configuration(self.configuration)
            else:
                raise ValueError("Couldn't save invalid configuration")

    def delete_configuration(self):
        """Delete the current configuration. Warning: it does not delete the file or resource of
        the configuration."""

        if self.configuration:
            if yesno("Save current configuration ?"):
                self.save_configuration()

            self.configuration = None

    def create_and_save_configuration_from_access_point(self):
        """Create and save a new configuration given data from access point."""
        # TODO: this function does not handle additional fitness protocols and efeatures yet.
        if self.access_point.has_targets_configuration():
            logger.info(
                "Targets configuration already present on access point."
                "Will not create another one."
            )
            return

        files = self.access_point.pipeline_settings.files_for_extraction
        targets = self.access_point.pipeline_settings.targets
        protocols_rheobase = self.access_point.pipeline_settings.protocols_rheobase
        protocols_mapping = self.access_point.pipeline_settings.protocols_mapping

        if not targets:
            auto_targets = self.access_point.pipeline_settings.auto_targets
            auto_targets_presets = self.access_point.pipeline_settings.auto_targets_presets
            if not auto_targets:
                if not auto_targets_presets:
                    raise TypeError(
                        "Please fill either targets, auto_targets or auto_targets_presets."
                        "Or alternatively save your targets before running your pipeline."
                    )
                auto_targets = get_auto_target_from_presets(auto_targets_presets)

        self.new_configuration(files, targets, protocols_rheobase, auto_targets, protocols_mapping)
        self.save_configuration()



"""Evaluator module."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import copy
import logging
import os
import pathlib

from bluepyopt.ephys.evaluators import CellEvaluator
from bluepyopt.ephys.locations import NrnSeclistCompLocation
from bluepyopt.ephys.locations import NrnSomaDistanceCompLocation
from bluepyopt.ephys.locations import NrnTrunkSomaDistanceCompLocation
from bluepyopt.ephys.objectives import SingletonWeightObjective
from bluepyopt.ephys.objectivescalculators import ObjectivesCalculator
from bluepyopt.ephys.simulators import NrnSimulator

from ..ecode import eCodes
from ..ecode import fixed_timestep_eCodes
from ..tools.multiprotocols_efeatures_utils import get_protocol_list_from_protocol_name
from ..tools.utils import are_same_protocol
from .efel_feature_bpem import DendFitFeature
from .efel_feature_bpem import DendFitMultiProtocolsFeature
from .efel_feature_bpem import eFELFeatureBPEM
from .protocols import BPEMProtocol
from .protocols import NoHoldingCurrent
from .protocols import ProtocolRunner
from .protocols import RinProtocol
from .protocols import RMPProtocol
from .protocols import SearchHoldingCurrent
from .protocols import SearchThresholdCurrent
from .protocols import ThresholdBasedProtocol
from .recordings import FixedDtRecordingCustom
from .recordings import LooseDtRecordingCustom

logger = logging.getLogger(__name__)

soma_loc = NrnSeclistCompLocation(name="soma", seclist_name="somatic", sec_index=0, comp_x=0.5)
ais_loc = NrnSeclistCompLocation(name="ais", seclist_name="axonal", sec_index=0, comp_x=0.5)

PRE_PROTOCOLS = [
    "SearchHoldingCurrent",
    "SearchThresholdCurrent",
    "RMPProtocol",
    "RinProtocol",
]
LEGACY_PRE_PROTOCOLS = ["RMP", "Rin", "RinHoldcurrent", "Main", "ThresholdDetection"]

seclist_to_sec = {
    "somatic": "soma",
    "apical": "apic",
    "axonal": "axon",
    "myelinated": "myelin",
}

protocol_type_to_class = {
    "Protocol": BPEMProtocol,
    "ThresholdBasedProtocol": ThresholdBasedProtocol,
    "ThresholdBasedNoHoldingProtocol": ThresholdBasedProtocol,
    "LocalThresholdBasedProtocol": ThresholdBasedProtocol,
}


def define_location(definition):
    # default case
    if definition is None or definition == "soma":
        return soma_loc

    if definition["type"] == "CompRecording":
        if definition["location"] == "soma":
            return soma_loc
        if definition["location"] == "ais":
            return ais_loc
        raise ValueError("Only soma and ais are implemented for CompRecording")

    if definition["type"] == "somadistance":
        return NrnSomaDistanceCompLocation(
            name=definition["name"],
            soma_distance=definition["somadistance"],
            seclist_name=definition["seclist_name"],
        )

    if definition["type"] == "somadistanceapic":
        return NrnTrunkSomaDistanceCompLocation(
            name=definition["name"],
            soma_distance=definition["somadistance"],
            seclist_name=definition["seclist_name"],
            direction="radial",
        )

    if definition["type"] == "nrnseclistcomp":
        return NrnSeclistCompLocation(
            name=definition["name"],
            comp_x=definition["comp_x"],
            sec_index=definition["sec_index"],
            seclist_name=definition["seclist_name"],
        )

    raise ValueError(f"Unknown location type {definition['type']}")


def define_recording(recording_conf, use_fixed_dt_recordings=False):
    """Create a recording from a configuration dictionary

    Args:
        recording_conf (dict): configuration of the recording. Must contain the type of the
            recording as well as information about the location of the recording (see function
            define_location).
        use_fixed_dt_recordings (bool): used for legacy currentscape
            to record at a fixed dt of 0.1 ms.

    Returns:
        FixedDtRecordingCustom or LooseDtRecordingCustom
    """

    location = define_location(recording_conf)
    variable = recording_conf.get("variable", recording_conf.get("var"))

    if use_fixed_dt_recordings:
        rec_class = FixedDtRecordingCustom
    else:
        rec_class = LooseDtRecordingCustom

    return rec_class(name=recording_conf["name"], location=location, variable=variable)


def define_protocol(
    protocol_configuration, stochasticity=False, use_fixed_dt_recordings=False, dt=None
):
    """Create a protocol.

    Args:
        protocol_configuration (ProtocolConfiguration): configuration of the protocol
        stochasticity (bool): Should the stochastic channels be stochastic or
            deterministic
        use_fixed_dt_recordings (bool): used for legacy currentscape to record at a fixed dt
            of 0.1 ms. However, Simulations will be run based on dt.
        dt (float): NEURON dt for fixed time step simulation. If None, variable
            dt will be used.
    Returns:
        Protocol
    """
    cvode_active = not dt

    recordings = []
    for rec_conf in protocol_configuration.recordings:
        recordings.append(define_recording(rec_conf, use_fixed_dt_recordings))

    if len(protocol_configuration.stimuli) != 1:
        raise ValueError("Only protocols with a single stimulus implemented")

    stim = protocol_configuration.stimuli[0]
    location_dict = stim.get("location", "soma")
    location = define_location(location_dict)
    # cannot use pop here because stim["location"] is used later
    filtered_stim = {k: v for k, v in stim.items() if k != "location"}
    for k, ecode in eCodes.items():
        if k in protocol_configuration.name.lower():
            stimulus = ecode(location=location, **filtered_stim)
            if k in fixed_timestep_eCodes:
                cvode_active = False
            break
    else:
        raise KeyError(
            f"There is no eCode linked to the stimulus name {protocol_configuration.name.lower()}. "
            "See ecode/__init__.py for the available stimuli "
            "names"
        )

    stoch = stochasticity and protocol_configuration.stochasticity
    if stoch:
        cvode_active = False

    kwargs = {}
    if protocol_configuration.protocol_type == "LocalThresholdBasedProtocol":
        loc_name = location_dict["name"]
        kwargs = {
            "hold_key": f"bpo_holding_current_{loc_name}",
            "thres_key": f"bpo_threshold_current_{loc_name}",
            "hold_prot_name": f"SearchHoldingCurrent_{loc_name}",
            "thres_prot_name": f"SearchThresholdCurrent_{loc_name}",
        }

    if protocol_configuration.protocol_type == "ThresholdBasedNoHoldingProtocol":
        kwargs = {
            "hold_key": "bpo_holding_current_noholding",
            "thres_key": "bpo_threshold_current_noholding",
            "hold_prot_name": "SearchHoldingCurrent_noholding",
            "thres_prot_name": "SearchThresholdCurrent_noholding",
        }

    if protocol_configuration.protocol_type in protocol_type_to_class:
        return protocol_type_to_class[protocol_configuration.protocol_type](
            name=protocol_configuration.name,
            stimulus=stimulus,
            recordings=recordings,
            cvode_active=cvode_active,
            stochasticity=stoch,
            **kwargs,
        )

    raise ValueError(f"Protocol type {protocol_configuration.protocol_type} not found")


def define_efeature(feature_config, protocol=None, global_efel_settings=None):
    """Define an efeature from a configuration dictionary"""

    global_efel_settings = {} if global_efel_settings is None else global_efel_settings

    stim_amp = None
    stim_start = None
    stim_end = None

    if protocol:
        stim_amp = protocol.amplitude

    if feature_config.efel_settings.get("stim_start", None) is not None:
        stim_start = feature_config.efel_settings["stim_start"]
    elif protocol:
        stim_start = protocol.stim_start()

    if feature_config.efel_settings.get("stim_end", None) is not None:
        stim_end = feature_config.efel_settings["stim_end"]
    elif protocol:
        if "bAP" in protocol.name:
            stim_end = protocol.total_duration
        else:
            stim_end = protocol.stim_end()

    efel_settings = {**global_efel_settings, **feature_config.efel_settings}

    # Handle the special case of multiple_decay_time_constant_after_stim
    if feature_config.efel_feature_name == "multiple_decay_time_constant_after_stim":
        if hasattr(protocol.stimuli[0], "multi_stim_start"):
            efel_settings["multi_stim_start"] = protocol.stimuli[0].multi_stim_start()
            efel_settings["multi_stim_end"] = protocol.stimuli[0].multi_stim_end()
        else:
            efel_settings["multi_stim_start"] = [stim_start]
            efel_settings["multi_stim_end"] = [stim_end]
    double_settings = {k: v for k, v in efel_settings.items() if isinstance(v, (float, list))}
    int_settings = {k: v for k, v in efel_settings.items() if isinstance(v, int)}
    string_settings = {k: v for k, v in efel_settings.items() if isinstance(v, str)}

    if "dendrite_backpropagation_fit" in feature_config.name:
        decay = "decay" in feature_config.name
        linear = "linear" in feature_config.name
        efeature = DendFitFeature(
            feature_config.name,
            efel_feature_name=feature_config.efel_feature_name,
            recording_names=feature_config.recording_name_for_instantiation,
            stim_start=stim_start,
            stim_end=stim_end,
            exp_mean=feature_config.mean,
            exp_std=feature_config.std,
            stimulus_current=stim_amp,
            threshold=efel_settings.get("Threshold", None),
            interp_step=efel_settings.get("interp_step", None),
            double_settings=double_settings,
            int_settings=int_settings,
            string_settings=string_settings,
            decay=decay,
            linear=linear,
            weight=feature_config.weight,
        )
    # protocol name contains list of location, e.g. 'apic[050,100,150]'
    elif "[" in feature_config.recording_name_for_instantiation[""]:
        decay = "decay" in feature_config.name
        linear = "linear" in feature_config.name
        efeature = DendFitMultiProtocolsFeature(
            feature_config.name,
            efel_feature_name=feature_config.efel_feature_name,
            recording_names=feature_config.recording_name_for_instantiation,
            stim_start=stim_start,
            stim_end=stim_end,
            exp_mean=feature_config.mean,
            exp_std=feature_config.std,
            stimulus_current=stim_amp,
            threshold=efel_settings.get("Threshold", None),
            interp_step=efel_settings.get("interp_step", None),
            double_settings=double_settings,
            int_settings=int_settings,
            string_settings=string_settings,
            decay=decay,
            linear=linear,
            weight=feature_config.weight,
        )
    else:
        efeature = eFELFeatureBPEM(
            feature_config.name,
            efel_feature_name=feature_config.efel_feature_name,
            recording_names=feature_config.recording_name_for_instantiation,
            stim_start=stim_start,
            stim_end=stim_end,
            exp_mean=feature_config.mean,
            exp_std=feature_config.std,
            stimulus_current=stim_amp,
            threshold=efel_settings.get("Threshold", None),
            interp_step=efel_settings.get("interp_step", None),
            double_settings=double_settings,
            int_settings=int_settings,
            string_settings=string_settings,
            weight=feature_config.weight,
        )

    return efeature


def define_RMP_protocol(
    efeatures,
    stimulus_duration=500.0,
    location=soma_loc,
    output_key="bpo_rmp",
    rmp_prot_name="RMPProtocol",
    recording_name=None,
):
    """Define the resting membrane potential protocol"""

    target_voltage = None
    for f in efeatures:
        if (
            "RMPProtocol" in f.recording_names[""]
            and f.efel_feature_name == "steady_state_voltage_stimend"
        ):
            if recording_name is not None:
                target_voltage = copy.deepcopy(f)
                target_voltage.recording_names = {"": f"{rmp_prot_name}.{recording_name}"}
            else:
                target_voltage = f
            break

    if not target_voltage:
        raise ValueError(
            "Couldn't find the efeature 'steady_state_voltage_stimend' associated to the "
            "'RMPProtocol' in your FitnessCalculatorConfiguration. It might not have been "
            "extracted from the ephys data you have available or the name of the protocol to"
            " use for RMP (setting 'name_rmp_protocol') might be wrong."
        )

    rmp_protocol = RMPProtocol(
        name=rmp_prot_name,
        location=location,
        target_voltage=target_voltage,
        stimulus_duration=stimulus_duration,
        output_key=output_key,
    )

    for f in efeatures:
        if (
            "RMPProtocol" in f.recording_names[""]
            and f.efel_feature_name != "steady_state_voltage_stimend"
        ):
            f.stim_start = 0.0
            f.stim_end = rmp_protocol.stimulus_duration
            f.stimulus_current = 0.0

    return rmp_protocol


def define_Rin_protocol(
    efeatures,
    amp=-0.02,
    stimulus_delay=500.0,
    stimulus_duration=500.0,
    totduration=1000.0,
    location=soma_loc,
    output_key="bpo_rin",
    hold_key="bpo_holding_current",
    hold_prot_name="SearchHoldingCurrent",
    rin_prot_name="RinProtocol",
    recording_name=None,
):
    """Define the input resistance protocol"""

    target_rin = None
    for f in efeatures:
        if (
            "RinProtocol" in f.recording_names[""]
            and f.efel_feature_name == "ohmic_input_resistance_vb_ssse"
        ):
            if recording_name is not None:
                target_rin = copy.deepcopy(f)
                target_rin.recording_names = {"": f"{rin_prot_name}.{recording_name}"}
            else:
                target_rin = f
            break

    if not target_rin:
        raise ValueError(
            "Couldn't find the efeature 'ohmic_input_resistance_vb_ssse' associated to "
            "the 'RinProtocol' in your FitnessCalculatorConfiguration. It might not have"
            "been extracted from the ephys data you have available or the name of the"
            " protocol to use for Rin (setting 'name_Rin_protocol') might be wrong."
        )

    return RinProtocol(
        name=rin_prot_name,
        location=location,
        target_rin=target_rin,
        amp=amp,
        stimulus_delay=stimulus_delay,
        stimulus_duration=stimulus_duration,
        totduration=totduration,
        output_key=output_key,
        hold_key=hold_key,
        hold_prot_name=hold_prot_name,
    )


def define_holding_protocol(
    efeatures,
    strict_bounds=False,
    max_depth=7,
    stimulus_duration=500.0,
    location=soma_loc,
    output_key="bpo_holding_current",
    hold_prot_name="SearchHoldingCurrent",
    recording_name=None,
):
    """Define the search holding current protocol"""

    target_voltage = None
    for f in efeatures:
        if (
            "SearchHoldingCurrent" in f.recording_names[""]
            and f.efel_feature_name == "steady_state_voltage_stimend"
        ):
            if recording_name is not None:
                target_voltage = copy.deepcopy(f)
                target_voltage.recording_names = {"": f"{hold_prot_name}.{recording_name}"}
            else:
                target_voltage = f
            break

    if target_voltage:
        return SearchHoldingCurrent(
            name=hold_prot_name,
            location=location,
            target_voltage=target_voltage,
            strict_bounds=strict_bounds,
            max_depth=max_depth,
            stimulus_duration=stimulus_duration,
            output_key=output_key,
        )

    raise ValueError(
        "Couldn't find the efeature 'bpo_holding_current' associated to "
        "the 'SearchHoldingCurrent' protocol in your FitnessCalculatorConfiguration."
    )


def define_threshold_protocol(
    efeatures,
    max_threshold_voltage=-30,
    step_delay=500.0,
    step_duration=2000.0,
    totduration=3000.0,
    spikecount_timeout=50,
    max_depth=10,
    location=soma_loc,
    no_spikes=True,
    efel_threshold=None,
    output_key="bpo_threshold_current",
    hold_key="bpo_holding_current",
    rmp_key="bpo_rmp",
    rin_key="bpo_rin",
    hold_prot_name="SearchHoldingCurrent",
    rmp_prot_name="RMPProtocol",
    rin_prot_name="RinProtocol",
    thres_prot_name="SearchThresholdCurrent",
):
    """Define the search threshold current protocol"""
    # pylint: disable=too-many-arguments

    target_current = []
    for f in efeatures:
        if thres_prot_name in f.recording_names[""] and f.efel_feature_name == output_key:
            target_current.append(f)

    if not target_current:
        target_current = [None]

    return SearchThresholdCurrent(
        name=thres_prot_name,
        location=location,
        target_threshold=target_current[0],
        max_threshold_voltage=max_threshold_voltage,
        stimulus_delay=step_delay,
        stimulus_duration=step_duration,
        stimulus_totduration=totduration,
        spikecount_timeout=spikecount_timeout,
        max_depth=max_depth,
        no_spikes=no_spikes,
        efel_threshold=efel_threshold,
        output_key=output_key,
        hold_key=hold_key,
        rmp_key=rmp_key,
        rin_key=rin_key,
        hold_prot_name=hold_prot_name,
        rmp_prot_name=rmp_prot_name,
        rin_prot_name=rin_prot_name,
    )


def define_protocols(
    fitness_calculator_configuration,
    include_validation_protocols,
    stochasticity,
    use_fixed_dt_recordings,
    dt,
):
    """Instantiate several efeatures"""

    protocols = {}
    for protocols_def in fitness_calculator_configuration.protocols:
        if not include_validation_protocols and protocols_def.validation:
            continue
        protocols[protocols_def.name] = define_protocol(
            protocols_def, stochasticity, use_fixed_dt_recordings, dt
        )

    return protocols


def define_efeatures(
    fitness_calculator_configuration,
    include_validation_protocols,
    protocols,
    efel_settings,
):
    """Instantiate several Protocols"""

    efeatures = []
    validation_prot = fitness_calculator_configuration.validation_protocols

    for feature_def in fitness_calculator_configuration.efeatures:
        if not include_validation_protocols and any(
            are_same_protocol(feature_def.protocol_name, p) for p in validation_prot
        ):
            continue

        protocol = None
        # split("_")[0] is to account for 'local' pre-protocol,
        # e.g. "SearchThresholdCurrent_apical055"
        if (
            feature_def.protocol_name not in PRE_PROTOCOLS
            and feature_def.protocol_name.split("_")[0] not in PRE_PROTOCOLS
        ):
            if "[" in feature_def.protocol_name:
                prot_names = get_protocol_list_from_protocol_name(feature_def.protocol_name)
            else:
                prot_names = [feature_def.protocol_name]
            for prot_name in prot_names:
                # here, we assume that all protocols used in a DendFitMultiProtocolsFeature
                # have the same stim_start and stim_end, and also that
                # we will not use any feature depending on protocol stimulus current.
                # As such, we can pass any protocol matching any item in prot_names.
                protocol = next((p for p in protocols.values() if p.name == prot_name), None)
                if protocol is None:
                    raise ValueError(f"Could not find protocol named {prot_name}")

        efeatures.append(define_efeature(feature_def, protocol, efel_settings))

    return efeatures


def define_optimisation_protocol(
    fitness_calculator_configuration,
    include_validation_protocols=False,
    stochasticity=True,
    efel_settings=None,
    use_fixed_dt_recordings=False,
    dt=None,
):
    """Create a meta protocol in charge of running the other protocols.

    Args:
        fitness_calculator_configuration (FitnessCalculatorConfiguration): configuration of the
            fitness calculator.
        include_validation_protocols (bool): should the validation protocols
            and validation efeatures be added to the evaluator.
        stochasticity (bool): Should the stochastic channels be stochastic or
            deterministic
        efel_settings (dict): eFEl settings.
        use_fixed_dt_recordings (bool): used for legacy currentscape
            to record at a fixed dt of 0.1 ms.
        dt (float): NEURON dt for fixed time step simulation. If None, variable dt will be used.
    """

    protocols = define_protocols(
        fitness_calculator_configuration,
        include_validation_protocols,
        stochasticity,
        use_fixed_dt_recordings,
        dt,
    )

    efeatures = define_efeatures(
        fitness_calculator_configuration,
        include_validation_protocols,
        protocols,
        efel_settings,
    )

    protocol_runner = ProtocolRunner(protocols)

    return protocol_runner, efeatures


def define_preprotocols(
    efeatures,
    location,
    fitness_calculator_configuration,
    strict_holding_bounds=False,
    max_depth_holding_search=7,
    max_threshold_voltage=-30,
    spikecount_timeout=50,
    max_depth_threshold_search=10,
    efel_thresh_for_thres_search=None,
    rmp_key="bpo_rmp",
    hold_key="bpo_holding_current",
    rin_key="bpo_rin",
    thres_key="bpo_threshold_current",
    rmp_prot_name="RMPProtocol",
    hold_prot_name="SearchHoldingCurrent",
    rin_prot_name="RinProtocol",
    thres_prot_name="SearchThresholdCurrent",
    recording_name=None,
    no_holding=False,
):
    """Returns a dict containing the preprotocols"""
    # pylint: disable=too-many-arguments
    rmp_prot = define_RMP_protocol(
        efeatures,
        stimulus_duration=fitness_calculator_configuration.rmp_duration,
        location=location,
        output_key=rmp_key,
        rmp_prot_name=rmp_prot_name,
        recording_name=recording_name,
    )
    if no_holding:
        holding_prot = NoHoldingCurrent(
            name=hold_prot_name,
            output_key=hold_key,
        )
        # without holding current, cell can spike spontaneously
        no_spikes = False
    else:
        holding_prot = define_holding_protocol(
            efeatures,
            strict_holding_bounds,
            max_depth_holding_search,
            stimulus_duration=fitness_calculator_configuration.search_holding_duration,
            location=location,
            output_key=hold_key,
            hold_prot_name=hold_prot_name,
            recording_name=recording_name,
        )
        no_spikes = True
    rin_prot = define_Rin_protocol(
        efeatures,
        amp=fitness_calculator_configuration.rin_step_amp,
        stimulus_delay=fitness_calculator_configuration.rin_step_delay,
        stimulus_duration=fitness_calculator_configuration.rin_step_duration,
        totduration=fitness_calculator_configuration.rin_totduration,
        location=location,
        output_key=rin_key,
        hold_key=hold_key,
        hold_prot_name=hold_prot_name,
        rin_prot_name=rin_prot_name,
        recording_name=recording_name,
    )
    threshold_prot = define_threshold_protocol(
        efeatures,
        max_threshold_voltage,
        fitness_calculator_configuration.search_threshold_step_delay,
        fitness_calculator_configuration.search_threshold_step_duration,
        fitness_calculator_configuration.search_threshold_totduration,
        spikecount_timeout,
        max_depth_threshold_search,
        location=location,
        no_spikes=no_spikes,
        efel_threshold=efel_thresh_for_thres_search,
        output_key=thres_key,
        hold_key=hold_key,
        rmp_key=rmp_key,
        rin_key=rin_key,
        hold_prot_name=hold_prot_name,
        rmp_prot_name=rmp_prot_name,
        rin_prot_name=rin_prot_name,
        thres_prot_name=thres_prot_name,
    )
    return {
        rmp_prot_name: rmp_prot,
        hold_prot_name: holding_prot,
        rin_prot_name: rin_prot,
        thres_prot_name: threshold_prot,
    }


def define_threshold_based_optimisation_protocol(
    fitness_calculator_configuration,
    include_validation_protocols=False,
    stochasticity=True,
    ais_recording=False,
    efel_settings=None,
    max_threshold_voltage=-30,
    strict_holding_bounds=True,
    use_fixed_dt_recordings=False,
    max_depth_holding_search=7,
    max_depth_threshold_search=10,
    spikecount_timeout=50,
    efel_threshold_for_threshold_search=None,
    dt=None,
):
    """Create a meta protocol in charge of running the other protocols.

    The amplitude of the "threshold_protocols" depend on the computation of
    the current threshold.

    Args:
        fitness_calculator_configuration (FitnessCalculatorConfiguration): configuration of the
            fitness calculator.
        include_validation_protocols (bool): should the validation protocols
            and validation efeatures be added to the evaluator.
        stochasticity (bool): Should the stochastic channels be stochastic or
            deterministic
        ais_recording (bool): if True all the soma recording will be at the first axonal section.
        efel_settings (dict): eFEl settings.
        max_threshold_voltage (float): maximum voltage at which the SearchThresholdProtocol
            will search for the rheobase.
        strict_holding_bounds (bool): to adaptively enlarge bounds if holding current is outside
            when set to False
        use_fixed_dt_recordings (bool): used for legacy currentscape
            to record at a fixed dt of 0.1 ms.
        max_depth_holding_search (int): maximum depth for the binary search for the
            holding current
        max_depth_threshold_search (int): maximum depth for the binary search for the
            threshold current
        spikecount_timeout (float): timeout for spikecount computation, if timeout is reached,
            we set spikecount=2 as if many spikes were present, to speed up bisection search.
        efel_threshold_for_threshold_search: spike threshold for the efel settings
            to use in SearchThresholdCurrent protocol. Set to None to keep the default value
            (currently -20 mV in efel)
        dt (float): NEURON dt for fixed time step simulation. If None, variable dt will be used.
    """

    protocols = define_protocols(
        fitness_calculator_configuration,
        include_validation_protocols,
        stochasticity,
        use_fixed_dt_recordings,
        dt,
    )

    efeatures = define_efeatures(
        fitness_calculator_configuration,
        include_validation_protocols,
        protocols,
        efel_settings,
    )

    # Create the special protocols for local stimulation
    for prot in fitness_calculator_configuration.protocols:
        if include_validation_protocols or not prot.validation:
            if prot.protocol_type == "LocalThresholdBasedProtocol":
                stim = prot.stimuli[0]
                loc_name = stim["location"]["name"]
                location = define_location(stim["location"])

                if f"RMPProtocol_{loc_name}" not in protocols:
                    protocols.update(
                        define_preprotocols(
                            efeatures=efeatures,
                            location=location,
                            fitness_calculator_configuration=fitness_calculator_configuration,
                            strict_holding_bounds=strict_holding_bounds,
                            max_depth_holding_search=max_depth_holding_search,
                            max_threshold_voltage=max_threshold_voltage,
                            spikecount_timeout=spikecount_timeout,
                            max_depth_threshold_search=max_depth_threshold_search,
                            efel_thresh_for_thres_search=efel_threshold_for_threshold_search,
                            rmp_key=f"bpo_rmp_{loc_name}",
                            hold_key=f"bpo_holding_current_{loc_name}",
                            rin_key=f"bpo_rin_{loc_name}",
                            thres_key=f"bpo_threshold_current_{loc_name}",
                            rmp_prot_name=f"RMPProtocol_{loc_name}",
                            hold_prot_name=f"SearchHoldingCurrent_{loc_name}",
                            rin_prot_name=f"RinProtocol_{loc_name}",
                            thres_prot_name=f"SearchThresholdCurrent_{loc_name}",
                            recording_name=f"{loc_name}.{stim['location']['variable']}",
                            # use holding current = 0 for local injection protocols
                            # TODO: let the user decide whether to use noholding or holding
                            no_holding=True,
                        )
                    )

            if prot.protocol_type == "ThresholdBasedNoHoldingProtocol":
                stim = prot.stimuli[0]
                location = ais_loc if ais_recording else soma_loc

                if "RMPProtocol_noholding" not in protocols:
                    protocols.update(
                        define_preprotocols(
                            efeatures=efeatures,
                            location=location,
                            fitness_calculator_configuration=fitness_calculator_configuration,
                            strict_holding_bounds=strict_holding_bounds,
                            max_depth_holding_search=max_depth_holding_search,
                            max_threshold_voltage=max_threshold_voltage,
                            spikecount_timeout=spikecount_timeout,
                            max_depth_threshold_search=max_depth_threshold_search,
                            efel_thresh_for_thres_search=efel_threshold_for_threshold_search,
                            rmp_key="bpo_rmp_noholding",
                            hold_key="bpo_holding_current_noholding",
                            rin_key="bpo_rin_noholding",
                            thres_key="bpo_threshold_current_noholding",
                            rmp_prot_name="RMPProtocol_noholding",
                            hold_prot_name="SearchHoldingCurrent_noholding",
                            rin_prot_name="RinProtocol_noholding",
                            thres_prot_name="SearchThresholdCurrent_noholding",
                            recording_name="soma.v",
                            no_holding=True,
                        )
                    )

            # Create the special protocols for soma stimulation
            if prot.protocol_type == "ThresholdBasedProtocol" and "RMPProtocol" not in protocols:
                location = ais_loc if ais_recording else soma_loc
                protocols.update(
                    define_preprotocols(
                        efeatures=efeatures,
                        location=location,
                        fitness_calculator_configuration=fitness_calculator_configuration,
                        strict_holding_bounds=strict_holding_bounds,
                        max_depth_holding_search=max_depth_holding_search,
                        max_threshold_voltage=max_threshold_voltage,
                        spikecount_timeout=spikecount_timeout,
                        max_depth_threshold_search=max_depth_threshold_search,
                        efel_thresh_for_thres_search=efel_threshold_for_threshold_search,
                        no_holding=False,
                    )
                )

    # Create the protocol runner
    protocol_runner = ProtocolRunner(protocols)

    return protocol_runner, efeatures


def define_fitness_calculator(features):
    """Creates the objectives calculator.

    Args:
        features (list): list of EFeature.

    Returns:
        ObjectivesCalculator
    """

    objectives = [SingletonWeightObjective(feat.name, feat, feat.weight) for feat in features]

    return ObjectivesCalculator(objectives)


def get_simulator(stochasticity, cell_model, dt=None, mechanisms_directory=None, cvode_minstep=0.0):
    """Get NrnSimulator

    Args:
        stochasticity (bool): allow the use of simulator for stochastic channels
        cell_model (CellModel): used to check if any stochastic channels are present
        dt (float): NEURON dt for fixed time step simulation. If None, variable dt will be used.
        mechanisms_directory (str or Path): path to the directory containing the mechanisms
        cvode_minstep (float): minimum time step allowed for a CVODE step.
    """

    if mechanisms_directory is not None:
        # To avoid double loading the same mechanisms:
        cwd = pathlib.Path(os.getcwd())
        mech_dir = pathlib.Path(mechanisms_directory)
        if cwd.resolve() == mech_dir.resolve():
            mechs_parent_dir = None
        else:
            mechs_parent_dir = str(mech_dir.parents[0])
    else:
        mechs_parent_dir = None

    if stochasticity:
        for mechanism in cell_model.mechanisms:
            if not mechanism.deterministic:
                return NrnSimulator(
                    dt=dt or 0.025,
                    cvode_active=False,
                    mechanisms_directory=mechs_parent_dir,
                )

        logger.warning(
            "Stochasticity is True but no mechanisms are stochastic. Switching to "
            "non-stochastic."
        )

    if dt is None:
        return NrnSimulator(mechanisms_directory=mechs_parent_dir, cvode_minstep=cvode_minstep)

    return NrnSimulator(dt=dt, cvode_active=False, mechanisms_directory=mechs_parent_dir)


def create_evaluator(
    cell_model,
    fitness_calculator_configuration,
    pipeline_settings,
    stochasticity=None,
    timeout=None,
    include_validation_protocols=False,
    mechanisms_directory=None,
    use_fixed_dt_recordings=False,
):
    """Creates an evaluator for a cell model/protocols/e-feature combo.

    Args:
        cell_model (CellModel): cell model
        fitness_calculator_configuration (FitnessCalculatorConfiguration): configuration of the
            fitness calculator.
        pipeline_settings (EModelPipelineSettings): settings for the pipeline.
        stochasticity (bool): should the stochastic channels be stochastic or
            deterministic
        timeout (float): maximum time in second during which a protocol is
            allowed to run
        include_validation_protocols (bool): should the validation protocols
            and validation efeatures be added to the evaluator.
        mechanisms_directory (str or Path): path to the directory containing the mechanisms.
        use_fixed_dt_recordings (bool): used for legacy currentscape
            to record at a fixed dt of 0.1 ms.

    Returns:
        CellEvaluator
    """

    stochasticity = pipeline_settings.stochasticity if stochasticity is None else stochasticity
    timeout = pipeline_settings.optimisation_timeout if timeout is None else timeout

    simulator = get_simulator(
        stochasticity=stochasticity,
        cell_model=cell_model,
        dt=pipeline_settings.neuron_dt,
        mechanisms_directory=mechanisms_directory,
        cvode_minstep=pipeline_settings.cvode_minstep,
    )

    fitness_calculator_configuration.configure_morphology_dependent_locations(cell_model, simulator)

    if any(
        p.protocol_type
        in (
            "ThresholdBasedProtocol",
            "LocalThresholdBasedProtocol",
            "ThresholdBasedNoHoldingProtocol",
        )
        for p in fitness_calculator_configuration.protocols
    ):
        main_protocol, features = define_threshold_based_optimisation_protocol(
            fitness_calculator_configuration,
            include_validation_protocols,
            stochasticity=stochasticity,
            efel_settings=pipeline_settings.efel_settings,
            max_threshold_voltage=pipeline_settings.max_threshold_voltage,
            strict_holding_bounds=pipeline_settings.strict_holding_bounds,
            use_fixed_dt_recordings=use_fixed_dt_recordings,
            max_depth_holding_search=pipeline_settings.max_depth_holding_search,
            max_depth_threshold_search=pipeline_settings.max_depth_threshold_search,
            spikecount_timeout=pipeline_settings.spikecount_timeout,
            efel_threshold_for_threshold_search=pipeline_settings.efel_settings.get(
                "Threshold", None
            ),
            dt=pipeline_settings.neuron_dt,
        )
    else:
        main_protocol, features = define_optimisation_protocol(
            fitness_calculator_configuration,
            include_validation_protocols,
            stochasticity=stochasticity,
            efel_settings=pipeline_settings.efel_settings,
            use_fixed_dt_recordings=use_fixed_dt_recordings,
            dt=pipeline_settings.neuron_dt,
        )

    fitness_calculator = define_fitness_calculator(features)
    fitness_protocols = {"main_protocol": main_protocol}

    param_names = [param.name for param in cell_model.params.values() if not param.frozen]

    cell_eval = CellEvaluator(
        cell_model=cell_model,
        param_names=param_names,
        fitness_protocols=fitness_protocols,
        fitness_calculator=fitness_calculator,
        sim=simulator,
        use_params_for_seed=pipeline_settings.use_params_for_seed,
        timeout=timeout,
    )
    cell_eval.prefix = cell_model.name

    return cell_eval


def add_recordings_to_evaluator(cell_evaluator, vars, use_fixed_dt_recordings=False):
    """Add a recording for each new variable for each protocol in cell evaluator."""
    # add recording for each protocol x new variable combination
    for prot in cell_evaluator.fitness_protocols["main_protocol"].protocols.values():
        if prot.name not in PRE_PROTOCOLS:
            base_rec = prot.recordings[0]
            for var in vars:
                location = base_rec.location

                split_name = base_rec.name.split(".")
                split_name[-1] = var
                name = ".".join(split_name)

                # FixedDtRecordingCustom for fixed time steps.
                # Use LooseDtRecordingCustom for variable time steps
                if use_fixed_dt_recordings:
                    new_rec = FixedDtRecordingCustom(name=name, location=location, variable=var)
                else:
                    new_rec = LooseDtRecordingCustom(name=name, location=location, variable=var)
                prot.recordings.append(new_rec)



"""Model Configurator"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from bluepyemodel.access_point.local import LocalAccessPoint
from bluepyemodel.model.neuron_model_configuration import NeuronModelConfiguration
from bluepyemodel.tools.utils import yesno

logger = logging.getLogger(__name__)


class ModelConfigurator:
    """Handles the loading, saving and modification of a model configuration"""

    def __init__(self, access_point, configuration=None):
        """Creates a model configuration, which includes the model parameters, distributions,
        mechanisms and a morphology.

        Args:
            access_point (DataAccessPoint): access point to the emodel data.
            configuration (NeuronModelConfiguration): a pre-existing configuration.
        """

        self.access_point = access_point
        self.configuration = configuration

    def new_configuration(self, use_gene_data=False):
        """Create a new configuration"""

        if self.configuration is not None:
            self.delete_configuration()

        if use_gene_data:
            self.get_gene_based_configuration()
        else:
            self.configuration = NeuronModelConfiguration(
                distributions=self.access_point.get_distributions(),
                available_mechanisms=self.access_point.get_available_mechanisms(),
                available_morphologies=self.access_point.get_available_morphologies(),
            )

    def load_configuration(self):
        """Load a previously registered configuration"""

        if isinstance(self.access_point, LocalAccessPoint):
            raise NotImplementedError(
                "Loading configuration is not yet implemented for local access point"
            )

        self.configuration = self.access_point.get_model_configuration()

    def save_configuration(self, path=None):
        """Save the configuration. The saving medium depends on the access point."""

        if self.configuration:
            self.access_point.store_model_configuration(self.configuration, path)

    def delete_configuration(self):
        """Delete the current configuration. Warning: it does not delete the file or resource of
        the configuration."""

        if self.configuration:
            if yesno("Save current configuration ?"):
                self.save_configuration()

            self.configuration = None

    def get_gene_based_parameters(self):
        """Get the gene mapping from Nexus and retrieve the matching parameters and mechanisms
        from the ion channel selector"""

        from bluepyemodel.icselector import icselector

        if not self.access_point.pipeline_settings.name_gene_map:
            logger.warning(
                "No gene mapping name informed. Only parameters registered by the user"
                " will be used."
            )
            return [], [], [], []

        _, gene_map_path = self.access_point.load_channel_gene_expression(
            self.access_point.pipeline_settings.name_gene_map
        )

        ic_map_path = self.access_point.load_ic_map()

        selector = icselector.ICSelector(ic_map_path, gene_map_path)
        me_type = f"{self.access_point.emodel_metadata.mtype}_ \
                    {self.access_point.emodel_metadata.etype}"
        parameters, mechanisms, distributions, nexus_keys = selector.get_cell_config_from_ttype(
            [self.access_point.emodel_metadata.ttype, me_type]
        )

        return parameters, mechanisms, distributions, nexus_keys

    def get_gene_based_configuration(self):
        """Overwrite the currently loaded configuration with a new configuration initiated from
        gene data."""

        self.configuration = NeuronModelConfiguration(
            distributions=self.access_point.get_distributions(),
            available_mechanisms=self.access_point.get_available_mechanisms(),
            available_morphologies=self.access_point.get_available_morphologies(),
        )

        params, mechs, distributions, nexus_keys = self.get_gene_based_parameters()

        for d in distributions:
            if d["name"] in ["uniform", "constant"]:
                continue
            function = d["function"] if "function" in d else d["fun"]
            self.configuration.add_distribution(
                d["name"],
                function,
                d.get("parameters", None),
                d.get("soma_ref_location", 0.5),
            )

        for p in params:
            self.configuration.add_parameter(
                p["name"],
                locations=p["location"],
                value=p["value"],
                mechanism=p.get("mechanism", "global"),
            )

        for m in mechs:
            version = None
            for k in nexus_keys:
                if k["name"] == m["name"]:
                    version = k.get("modelid") if "modelid" in k else k.get("modelId", None)

            temp_entry = m.get("temperature", None)
            if isinstance(temp_entry, dict) and "value" in temp_entry:
                temperature = temp_entry.get("value")
            else:
                temperature = temp_entry

            self.configuration.add_mechanism(
                m["name"],
                locations=m["location"],
                stochastic=m.get("stochastic", None),
                version=version,
                temperature=temperature,
                ljp_corrected=(
                    m.get("ljp_corrected")
                    if "ljp_corrected" in m
                    else m.get("isLjpCorrected", None)
                ),
            )



"""Utility module for evaluation."""

"""
Copyright 2023-2024, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from ..ecode import eCodes
from .efel_feature_bpem import DendFitFeature
from .evaluator import define_location
from .evaluator import protocol_type_to_class
from .recordings import LooseDtRecordingCustom

logger = logging.getLogger(__name__)


def add_dendritic_recordings(
    recordings, prot_name, dend_type, dist_start=10, dist_end=600, dist_step=10
):
    """Add dendritic recordings to recordings list.

    Args:
        recordings (list): variable to be updated with new recordings
        prot_name (str): protocol name
        dend_type (str): On which dendrite should the recording be.
            Can be 'apical' or 'basal'
        dist_start (int): closest distance on the dendrite on which to record
        dist_end (int): furthest distance on the dendrite on which to record
        dist_step (int): record every dist_step distance on the dendrite
    """
    for dist in range(dist_start, dist_end, dist_step):
        rec_dict = {
            "type": "somadistanceapic",
            "somadistance": dist,
            "name": f"{prot_name}.{dend_type}{dist:03d}.v",
            "seclist_name": dend_type,
        }
        new_loc = define_location(rec_dict)
        rec = LooseDtRecordingCustom(name=rec_dict["name"], location=new_loc, variable="v")
        recordings.append(rec)


def define_bAP_protocol(dist_start=10, dist_end=600, dist_step=10, dist_end_basal=150):
    """Utility function to create a ready-to-use bAP protocol.

    Args:
        dist_start (int): closest distance on the dendrites on which to record
        dist_end (int): furthest distance on the apical dendrite on which to record
        dist_step (int): record every dist_step distance on the dendrites
        dist_end_basal (int): furthest distance on the basal dendrite on which to record
    """
    name = "bAP_1000"
    soma_loc = define_location("soma")
    stim = {
        "delay": 700,
        "amp": None,
        "thresh_perc": 1000.0,
        "duration": 4.0,
        "totduration": 1000,
        "holding_current": 0,
    }
    stimulus = eCodes["idrest"](location=soma_loc, **stim)

    recordings = [LooseDtRecordingCustom(name=f"{name}.soma.v", location=soma_loc, variable="v")]
    add_dendritic_recordings(recordings, name, "apical", dist_start, dist_end, dist_step)
    add_dendritic_recordings(recordings, name, "basal", dist_start, dist_end_basal, dist_step)

    return protocol_type_to_class["ThresholdBasedProtocol"](
        name=name,
        stimulus=stimulus,
        recordings=recordings,
        cvode_active=True,
        stochasticity=False,
    )


def define_bAP_feature(dend_type="apical", dist_start=10, dist_end=600, dist_step=10):
    """Utility function to create a ready-to-use dendrite backpropagation fit feature

    Args:
        dend_type (str): Which dendrite data should the feature compute.
            Can be 'apical' or 'basal'
        dist_start (int): closest distance on the dendrite on which to record
        dist_end (int): furthest distance on the dendrite on which to record
        dist_step (int): record every dist_step distance on the dendrite
    """
    name = "bAP_1000"

    recording_names = {"": f"{name}.soma.v"}
    for dist in range(dist_start, dist_end, dist_step):
        recording_names[dist] = f"{name}.{dend_type}{dist:03d}.v"

    feat = DendFitFeature(
        f"{dend_type}_dendrite_backpropagation_fit_decay",
        efel_feature_name="maximum_voltage_from_voltagebase",
        recording_names=recording_names,
        stim_start=700.0,
        stim_end=704.0,
        exp_mean=765.2223254692782,  # filler
        exp_std=37.98402943833677,  # filler
        stimulus_current=0.0,  # filler
        threshold=-30.0,
        interp_step=0.025,
        double_settings={},
        int_settings={"strict_stiminterval": 1},
        string_settings={},
        decay=True,
        linear=False,
    )
    return feat


def define_EPSP_protocol(dend_type, dist_start=100, dist_end=600, dist_step=100):
    """Returns ready-to-use EPSP protocols at multiple locations along the dendrite.

    Args:
        dend_type (str): Which dendrite data should the feature compute.
            Can be 'apical' or 'basal'
        dist_start (int): closest distance on the dendrite on which to record
        dist_end (int): furthest distance on the dendrite on which to record
        dist_step (int): record every dist_step distance on the dendrite
    """
    prots = {}
    # should we translate apical to apic here or not?
    soma_loc = define_location("soma")
    stim = {
        "syn_weight": 1.13,
        "syn_delay": 400.0,
        "totduration": 500.0,
    }

    # create a protocol injecting at soma location to have a datapoint at 0 distance
    name = "ProbAMPANMDA_EMS_0"
    stimulus = eCodes["probampanmda_ems"](location=soma_loc, **stim)
    recordings = [LooseDtRecordingCustom(name=f"{name}.soma.v", location=soma_loc, variable="v")]
    prot = protocol_type_to_class["Protocol"](
        name=name,
        stimulus=stimulus,
        recordings=recordings,
        cvode_active=False,  # cannot be used with cvode
        stochasticity=False,
    )
    prots[prot.name] = prot

    # create protocols injecting in the dendrites
    for dist in range(dist_start, dist_end, dist_step):
        name = f"ProbAMPANMDA_EMS{dend_type}{dist:03d}_0"
        loc_name = f"{dend_type}{dist:03d}"
        loc = define_location(
            {
                "type": "somadistanceapic",
                "somadistance": dist,
                "name": loc_name,
                "seclist_name": dend_type,
            }
        )
        stimulus = eCodes["probampanmda_ems"](location=loc, **stim)

        recordings = [
            LooseDtRecordingCustom(name=f"{name}.{loc_name}.v", location=loc, variable="v"),
            LooseDtRecordingCustom(name=f"{name}.soma.v", location=soma_loc, variable="v"),
        ]

        prot = protocol_type_to_class["Protocol"](
            name=name,
            stimulus=stimulus,
            recordings=recordings,
            cvode_active=False,  # cannot be used with cvode
            stochasticity=False,
        )
        prots[prot.name] = prot

    return prots


def define_EPSP_feature(
    dend_type="apical", rec_loc="dend", dist_start=100, dist_end=600, dist_step=100
):
    """Utility function to create a ready-to-use dendrite EPSP fit feature

    Args:
        dend_type (str): Which dendrite data should the feature compute.
            Can be 'apical' or 'basal'
        rec_loc (str): where should data be recorded from.
            Can be either 'dend' or 'soma'
        dist_start (int): closest distance on the dendrite on which to record
        dist_end (int): furthest distance on the dendrite on which to record
        dist_step (int): record every dist_step distance on the dendrite

    Attention! This is just used for plotting function,
    BPEM cannot currently handle EPSP attenuation fit feature on its own
    """
    recording_names = {"": "ProbAMPANMDA_EMS_0.soma.v"}
    for dist in range(dist_start, dist_end, dist_step):
        loc_name = f"{dend_type}{dist:03d}"
        if rec_loc == "dend":
            rec_name = loc_name
        else:
            rec_name = "soma"
        recording_names[dist] = f"ProbAMPANMDA_EMS{loc_name}_0.{rec_name}.v"

    feat = DendFitFeature(
        f"EPSP_{dend_type}_{rec_loc}",
        efel_feature_name="maximum_voltage_from_voltagebase",
        recording_names=recording_names,
        stim_start=400.0,
        stim_end=500.0,
        exp_mean=1.0,  # filler
        exp_std=1.0,  # filler
        stimulus_current=0.0,  # filler
        threshold=-30.0,
        interp_step=0.025,
        double_settings={},
        int_settings={"strict_stiminterval": 1},
        string_settings={},
        decay=False,
        linear=False,
    )
    return feat
